import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/keytab"
requires_modules = ['wga']
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve Kerberos: Keytab
    """
    return isamAppliance.invoke_get("Retrieve Kerberos: Keytab", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)


def _check(isamAppliance, id):
    """
    Check if keytab file is present

    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance)

    logger.debug("Looking for {0} existing keytab files in: {1}".format(id, ret_obj['data']))
    if ret_obj['data']:
        for keytab in ret_obj['data']:
            if keytab['id'] == id:
                logger.debug("Found keytab: {0}".format(id))
                return True

    return False


def import_keytab(isamAppliance, id, file, check_mode=False, force=False):
    """
    Import a keytab file, id is the name of the keytab file (they should be the same)
    """
    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(description="Import a keytab file",
                                                   uri="{0}".format(uri),
                                                   fileinfo=[{
                                                       'file_formfield': 'keytab_file',
                                                       'filename': file,
                                                       'mimetype': 'application/octet-stream'
                                                   }],
                                                   data={},
                                                   requires_modules=requires_modules, requires_version=requires_version,
                                                   json_response=False)

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Delete a keytab file, id is the name of the keytab file
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(description="Delete a keytab file", uri="{0}/{1}".format(uri, id),
                                               requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def combine(isamAppliance, newname, keytab_files, check_mode=False, force=False):
    """
    Combine two or more keytab files into one

    keytab_files should be an array of keytabs, like so:
        ['a.keytab', 'b.keytab']
    """
    warnings = []
    for keytab in keytab_files:
        if _check(isamAppliance, keytab) is False:
            warnings.append("keytab file to be combined: {0}, not found".format(keytab))

    if force is True or _check(isamAppliance, newname) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(description="Combine keytab files", uri=uri,
                                            data={
                                                "new_name": newname,
                                                "keytab_files": keytab_files
                                            }, warnings=warnings,
                                            requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare kerberos keytab files between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import ibmsecurity.utilities.tools
import time

logger = logging.getLogger(__name__)

requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current FIPS Mode configuration
    """
    return isamAppliance.invoke_get("Retrieving the current FIPS Mode configuration",
                                    "/fips_cfg", requires_model=requires_model)


def set(isamAppliance, fipsEnabled=True, tlsv10Enabled=True, tlsv11Enabled=False, check_mode=False, force=False):
    """
    Updating the FIPS Mode configuration
    """

    obj = _check(isamAppliance, fipsEnabled, tlsv10Enabled, tlsv11Enabled)
    if force is True or obj['value'] is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=obj['warnings'])
        else:
            return isamAppliance.invoke_put(
                "Updating the FIPS Mode configuration",
                "/fips_cfg",
                {
                    "fipsEnabled": fipsEnabled,
                    "tlsv10Enabled": tlsv10Enabled,
                    "tlsv11Enabled": tlsv11Enabled
                },
                requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=obj['warnings'])


def restart(isamAppliance, check_mode=False, force=False):
    """
    Rebooting and enabling the FIPS Mode configuration
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Rebooting and enabling the FIPS Mode configuration",
            "/fips_cfg/restart",
            {}, requires_model=requires_model
        )


def restart_and_wait(isamAppliance, wait_time=300, check_freq=5, check_mode=False, force=False):
    """
    Restart after FIPS configuration changes
    :param isamAppliance:
    :param wait_time:
    :param check_freq:
    :param check_mode:
    :param force:
    :return:
    """

    if isamAppliance.facts['model'] != "Appliance":
        return isamAppliance.create_return_object(
            warnings="API invoked requires model: {0}, appliance is of deployment model: {1}.".format(
                requires_model, isamAppliance.facts['model']))

    warnings = []
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        firmware = ibmsecurity.isam.base.firmware.get(isamAppliance, check_mode=check_mode, force=force)

        ret_obj = restart(isamAppliance)

        if ret_obj['rc'] == 0:
            sec = 0

            # Now check if it is up and running
            while 1:
                ret_obj = ibmsecurity.isam.base.firmware.get(isamAppliance, check_mode=check_mode, force=force,
                                                             ignore_error=True)

                # check partition last_boot time
                if ret_obj['rc'] == 0 and isinstance(ret_obj['data'], list) and len(ret_obj['data']) > 0 and \
                        (('last_boot' in ret_obj['data'][0] and ret_obj['data'][0]['last_boot'] != firmware['data'][0][
                            'last_boot'] and ret_obj['data'][0]['active'] == True) or (
                                 'last_boot' in ret_obj['data'][1] and ret_obj['data'][1]['last_boot'] !=
                                 firmware['data'][1]['last_boot'] and ret_obj['data'][1]['active'] == True)):
                    logger.info("Server is responding and has a different boot time!")
                    return isamAppliance.create_return_object(warnings=warnings)
                else:
                    time.sleep(check_freq)
                    sec += check_freq
                    logger.debug(
                        "Server is not responding yet. Waited for {0} secs, next check in {1} secs.".format(sec,
                                                                                                            check_freq))

                if sec >= wait_time:
                    warnings.append(
                        "The FIPS restart not detected or completed, exiting... after {0} seconds".format(sec))
                    break

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, fipsEnabled, tlsv10Enabled, tlsv11Enabled):
    obj = {'value': True, 'warnings': ""}

    ret_obj = get(isamAppliance)
    obj['warnings'] = ret_obj['warnings']

    if ret_obj['data']['fipsEnabled'] != fipsEnabled:
        logger.info("fipsEnabled change to {0}".format(fipsEnabled))
        obj['value'] = False
        return obj

    if ret_obj['data']['tlsv10Enabled'] != tlsv10Enabled:
        logger.info("TLS v1.0 change to {0}".format(tlsv10Enabled))
        obj['value'] = False
        return obj

    if ret_obj['data']['tlsv11Enabled'] != tlsv11Enabled:
        logger.info("TLS v1.1 change to {0}".format(tlsv11Enabled))
        obj['value'] = False
        return obj

    return obj


def compare(isamAppliance1, isamAppliance2):
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import json
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import logging
from .ibmappliance import IBMAppliance
from .ibmappliance import IBMError
from .ibmappliance import IBMFatal
from ibmsecurity.utilities import tools
from io import open
from os import environ

try:
    basestring
except NameError:
    basestring = (str, bytes)


class ISVGAppliance(IBMAppliance):
    def __init__(self, hostname, user, lmi_port=443, verify=None):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating an ISVGAppliance')
        if isinstance(lmi_port, basestring):
            self.lmi_port = int(lmi_port)
        else:
            self.lmi_port = lmi_port
        self.hostname = hostname
        self.session = requests.session()

        # If we did not get a value for verify, try the environment variable
        if verify is None:
            verify = str(environ.get("IBMSECLIB_VERIFY_CONNECTION", False)).lower() in ["true", "yes"]

        self.cert = cert

        self.disable_urllib_warnings = False
        if self.cert is None:
            self.logger.debug('Cert object is None, using BA Auth with userid/password.')
            self.session.auth = (user.username, user.password)
        else:
            self.logger.debug('Using cert based auth, since cert object is not None.')
            self.session.cert = self.cert

        self._set_ssl_verification(requests_verify_param=verify)

        IBMAppliance.__init__(self, hostname, user)

    def _set_ssl_verification(self, requests_verify_param):
        self.verify = requests_verify_param
        self.session.verify = self.verify
        if self.verify is None or self.verify is False:
            self.disable_urllib_warnings = True
            self.logger.warning("""
Certificate verification has been disabled. Python is NOT verifying the SSL 
certificate of the host appliance and InsecureRequestWarning messages are 
being suppressed for the following host:
  https://{0}:{1}

To use certificate verification:
  1. When the certificate is trusted by your Python environment:
        Instantiate all instances of ISVGAppliance with verify=True or set 
        the environment variable IBMSECLIB_VERIFY_CONNECTION=True.
  2. When the certificate is not already trusted in your Python environment:
        Instantiate all instances of ISAMAppliance with the verify parameter
        set to the fully qualified path to a CA bundle.
        
See the following URL for more details:
  https://requests.readthedocs.io/en/latest/user/advanced/#ssl-cert-verification
""".format(self.hostname, self.lmi_port))

    def _url(self, uri):
        # Build up the URL
        url = "https://" + self.hostname + ":" + str(self.lmi_port) + uri
        self.logger.debug("Issuing request to: " + url)

        return url

    def _log_desc(self, description):
        if description != "":
            self.logger.info('*** ' + description + ' ***')

    def _suppress_ssl_warning(self):
        # If we have trust setup correctly, we do not want to suppress these warnings.
        if not self.disable_urllib_warnings:
            return

        # Disable https warning because of non-standard certs on appliance
        try:
            self.logger.warning("Suppressing SSL Warnings.")
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        except AttributeError:
            self.logger.warning("load requests.packages.urllib3.disable_warnings() failed")

    def _process_response(self, return_obj, http_response, ignore_error):

        return_obj['rsp'] = http_response
        return_obj['rc'] = http_response.status_code

        # Examine the response.
        if (http_response.status_code == 302):
            self.logger.warning("  Endpoint cannot be used at this time: ")
            self.logger.warning("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            # Too early to use this endpoint.
            return_obj['changed'] = False
            return_obj['rc'] = 0
        elif (http_response.status_code != 200 and http_response.status_code != 204 and http_response.status_code != 201):
            self.logger.error("  Request failed: ")
            self.logger.error("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            if not ignore_error:
                raise IBMError("HTTP Return code: {0}".format(http_response.status_code), http_response.text)
            return_obj['changed'] = False  # force changed to be False as there is an error
        else:
            return_obj['rc'] = 0

        # Handle if there was json on input but response was not in json format
        json_data = {}
        try:
            json_data = json.loads(http_response.text)
        except ValueError:
            return_obj['data'] = http_response.content
            return

        self.logger.debug("Status Code: {0}".format(http_response.status_code))
        if http_response.text != "":
            self.logger.debug("Text: " + http_response.content.decode("utf-8"))

        for key in http_response.headers:
            if key == 'g-type':
                if http_response.headers[key] == 'application/octet-stream; charset=UTF-8':
                    json_data = {}
                    return_obj.data = http_response.content
                    return

        if http_response.text != "":
            json_data = json.loads(http_response.text)

        return_obj['data'] = json_data

    def _process_connection_error(self, ignore_error, return_obj):
        if not ignore_error:
            self.logger.critical("Failed to connect to server.")
            raise IBMError("HTTP Return code: 502", "Failed to connect to server")
        else:
            self.logger.debug("Failed to connect to server.")
            return_obj['rc'] = 502

    def _process_warnings(self, uri, requires_modules, requires_version, warnings=[]):
        # flag to indicate if processing needs to return and not continue
        return_call = False
        self.logger.debug("Checking for minimum version: {0}.".format(requires_version))
        if requires_version is not None and 'version' in self.facts and self.facts['version'] is not None:
            if self.facts['version'] < requires_version:
                return_call = True
                warnings.append(
                    "API invoked requires minimum version: {0}, appliance is of lower version: {1}.".format(
                        requires_version, self.facts['version']))
        # Detecting modules from uri if none is provided
        if requires_modules is None and not requires_modules:
            if uri.startswith("/wga"):
                requires_modules = ['wga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))
            elif uri.startswith("/mga"):
                requires_modules = ['mga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))

        self.logger.debug("Checking for one of required modules: {0}.".format(requires_modules))
        if requires_modules is not None and requires_modules:
            if 'activations' in self.facts and self.facts['activations']:
                # Find intersection of the two lists
                iactive = [ia for ia in self.facts['activations'] if ia in requires_modules]
                if not iactive:
                    return_call = True
                    warnings.append(
                        "API invoked requires one of modules: {0}, appliance has these modules active: {1}.".format(
                            requires_modules, self.facts['activations']))
                else:
                    self.logger.info("Modules satisfying requirement: {0}".format(iactive))
            else:
                return_call = True
                warnings.append("API invoked requires module: {0}, appliance has no modules active.".format(
                    requires_modules))

        self.logger.debug("Warnings: {0}".format(warnings))
        return warnings, return_call

    def invoke_post_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                          requires_version=None, warnings=[], json_response=True):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        if json_response:
            headers = {
                'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
            }
        else:
            headers = {
                'Accept': 'text/html,application/xhtml+xml,application/xml'
            }
        self.logger.debug("Headers are: {0}".format(headers))

        files = list()
        for file2post in fileinfo:
            files.append((file2post['file_formfield'],
                          (tools.path_leaf(file2post['filename']), open(file2post['filename'], 'rb'),
                           file2post['mimetype'])))

        self._suppress_ssl_warning()

        try:
            r = requests.post(url=self._url(uri=uri), data=data, auth=(self.user.username, self.user.password),
                              files=files, verify=self.verify, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_put_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                         requires_version=None, warnings=[]):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        headers = {
            'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        files = list()

        for file2post in fileinfo:
            files.append((file2post['file_formfield'],
                          (file2post['filename'], open(file2post['filename'], 'rb'), file2post['mimetype'])))

        self._suppress_ssl_warning()

        try:
            r = requests.put(url=self._url(uri=uri), data=data, auth=(self.user.username, self.user.password),
                             files=files, verify=self.verify, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_get_file(self, description, uri, filename, no_headers=False, mime_types=None, ignore_error=False, requires_modules=None,
                        requires_version=None, warnings=[]):
        """
        Invoke a GET request and download the response data to a file
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # In some cases passing a header causes response to come back as JSON
        if no_headers is True:
            headers = {}
        elif mime_types is None:
            headers = {
                'Accept': 'application/json,application/octet-stream'
            }
        else:
            headers = {
                'Accept': mime_types
            }
        self.logger.debug("Headers are: {0}".format(headers))

        self._suppress_ssl_warning()

        try:
            r = requests.get(url=self._url(uri=uri), auth=(self.user.username, self.user.password), verify=self.verify,
                             stream=True, headers=headers, allow_redirects=False)

            if (r.status_code != 200 and r.status_code != 204 and r.status_code != 201):
                self.logger.error("  Request failed: ")
                self.logger.error("     status code: {0}".format(r.status_code))
                if r.text != "":
                    self.logger.error("     text: " + r.text)
                if not ignore_error:
                    raise IBMError("HTTP Return code: {0}".format(r.status_code), r.text)
                else:
                    return_obj['rc'] = r.status_code
                    return_obj['data'] = {'msg': 'Unable to extract contents to file!'}
            else:
                with open(filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                return_obj['rc'] = 0
                return_obj['data'] = {'msg': 'Contents extracted to file: ' + filename}

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        except IOError:
            if not ignore_error:
                self.logger.critical("Failed to write to file: " + filename)
                raise IBMError("HTTP Return code: 999", "Failed to write to file: " + filename)
            else:
                self.logger.debug("Failed to write to file: " + filename)
                return_obj['rc'] = 999

        return return_obj

    def _invoke_request(self, func, description, uri, ignore_error, data={}, requires_modules=None,
                        requires_version=None, warnings=[]):
        """
        Send a request to the LMI.  This function is private and should not be
        used directly.  The invoke_get/invoke_put/etc functions should be used instead.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # There maybe some cases when header should be blank (not json)
        headers = {
            'Accept': 'application/json',
            'Content-type': 'application/json'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        # Process the input data into JSON
        json_data = json.dumps(data)

        self.logger.debug("Input Data: " + json_data)

        self._suppress_ssl_warning()

        try:
            if func == requests.get or func == requests.delete:

                if data != {}:
                    r = func(url=self._url(uri), data=json_data, auth=(self.user.username, self.user.password),
                             verify=self.verify, headers=headers, allow_redirects=False)
                else:
                    r = func(url=self._url(uri), auth=(self.user.username, self.user.password),
                             verify=self.verify, headers=headers, allow_redirects=False)
            else:
                r = func(url=self._url(uri), data=json_data,
                         auth=(self.user.username, self.user.password),
                         verify=self.verify, headers=headers)

            if func != requests.get:
                return_obj['changed'] = True  # Anything but GET should result in change

            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        return return_obj

    def invoke_put(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[]):
        """
        Send a PUT request to the LMI.
        """
        return self._invoke_request(requests.put, description, uri, ignore_error, data,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    warnings=warnings)

    def invoke_post(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                    warnings=[]):
        """
        Send a POST request to the LMI.
        """
        return self._invoke_request(requests.post, description, uri, ignore_error, data,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    warnings=warnings)

    def invoke_get(self, description, uri, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[]):
        """
        Send a GET request to the LMI.
        """
        return self._invoke_request(requests.get, description, uri, ignore_error, requires_modules=requires_modules,
                                    requires_version=requires_version, warnings=warnings)

    def invoke_delete(self, description, uri, ignore_error=False, requires_modules=None, requires_version=None,
                      warnings=[]):
        """
        Send a DELETE request to the LMI.
        """
        return self._invoke_request(requests.delete, description, uri, ignore_error, requires_modules=requires_modules,
                                    requires_version=requires_version, warnings=warnings)

    def get_facts(self):
        """
        Get facts about the appliance
        """
        # Fact collection will abort on any exception
        try:
            self.get_version()
        # Exceptions like those connection related will be ignored
        except:
            pass

    def get_version(self):
        """
        Get  appliance version (active partition)

        When firmware are installed or partition are changed, then this value is updated
        """
        self.facts['version'] = None
        import ibmsecurity.isvg.firmware

        ret_obj = ibmsecurity.isvg.firmware.get(self)
        for partition in ret_obj['data']:
            if partition['active'] is True:
                ver = partition['firmware_version'].split(' ')
                self.facts['version'] = ver[-1]



import logging
from ibmsecurity.utilities import tools
import os.path

logger = logging.getLogger(__name__)

uri = "/support"
requires_modules = None
requires_version = "9.0.2.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving information about all valid support files
    """
    return isamAppliance.invoke_get("Retrieving information about all valid support files", uri)


def get_categories(isamAppliance, check_mode=False, force=False):
    """
    Retrieving information about all the support categories
    """
    return isamAppliance.invoke_get("Retrieving information about all the support categories",
                                    "{0}/categories".format(uri))


def get_instances(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieving information about all the instances for a specific support category
    """
    return isamAppliance.invoke_get("Retrieving information about all the instances for a specific support category",
                                    "{0}/categories/{1}/instances".format(uri, name))


def create(isamAppliance, comment='', wrp=None, isam_runtime=None, lmi=None, cluster=None, felb=None,
           aac_federation=None, system=None, check_mode=False, force=False):
    """
    Creating a new support file
    """
    json_data = {}
    json_data['comment'] = comment
    min_version = True

    if 'version' in isamAppliance.facts and isamAppliance.facts['version'] is not None:
        if tools.version_compare(isamAppliance.facts['version'], requires_version) < 0:
            min_version = False

    if min_version is True:
        if wrp != None:
            json_data['wrp'] = wrp

        if isam_runtime != None:
            json_data['isam_runtime'] = isam_runtime

        if lmi != None:
            json_data['lmi'] = lmi

        if cluster != None:
            json_data['cluster'] = cluster

        if felb != None:
            json_data['felb'] = felb

        if aac_federation != None:
            json_data['aac_federation'] = aac_federation

        if system != None:
            json_data['system'] = system

    if force is True or _check(isamAppliance, comment=comment) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Creating a new support file",
                                             "{0}/".format(uri),
                                             json_data,
                                             requires_modules=requires_modules,
                                             requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, comment='', id=None):
    """
    Check if the last created support file has the exact same comment or id exists

    :param isamAppliance:
    :param comment:
    :return:
    """
    ret_obj = get(isamAppliance)

    if id != None:
        for sups in ret_obj['data']:
            if sups['id'] == id:
                return True
    else:
        for sups in ret_obj['data']:
            if sups['comment'] == comment:
                return True

    #        if isinstance(ret_obj['data'], list):
    #            # check only latest comment
    #            sup_file = min(ret_obj['data'], key=lambda sup: sup['index'])
    #            if sup_file['comment'] == comment:
    #                return True

    return False


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a support file
    """
    if force is True or _check(isamAppliance, id=id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if isinstance(id, list):
                del_uri = "{0}/multi_destroy{1}".format(uri, tools.create_query_string(record_ids=','.join(id)))
            else:
                del_uri = "{0}/{1}".format(uri, id)
            return isamAppliance.invoke_delete("Deleting a support file", del_uri)

    return isamAppliance.create_return_object()


def modify(isamAppliance, id, filename, comment, check_mode=False, force=False):
    """
    Modify the snapshot comment
    """
    if force is True or _check(isamAppliance, id=id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Modifying snapshot", "{0}/{1}".format(uri, id),
                                            {
                                                'id': id,
                                                'filename': filename,
                                                'comment': comment
                                            })

    return isamAppliance.create_return_object()


def download(isamAppliance, filename, id, check_mode=False, force=False):
    """
    Download snapshot file(s) to a zip file.
    Note: id can be a list or a single value
    """
    if force is True or (_check(isamAppliance, id=id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            if isinstance(id, list):
                id = ','.join(id)
            uri_download = "{0}/download{1}".format(uri, tools.create_query_string(record_ids=id))
            return isamAppliance.invoke_get_file("Downloading snapshots", uri_download, filename)

    return isamAppliance.create_return_object()


def download_latest(isamAppliance, dir='.', check_mode=False, force=False):
    """
    Download latest support file to a zip file.
    """
    ret_obj = get(isamAppliance)

    # Get snapshot with lowest 'id' value - that will be latest one
    sup_file = min(ret_obj['data'], key=lambda sup: sup['index'])
    id = sup_file['id']
    file = sup_file['filename']
    filename = os.path.join(dir, file)

    return download(isamAppliance, filename, id, check_mode, force)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare list of support files between 2 appliances - not sure if this will ever be useful?
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for snapshot in ret_obj1['data']:
        del snapshot['id']
        del snapshot['filename']

    for snapshot in ret_obj2['data']:
        del snapshot['id']
        del snapshot['filename']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'filename'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/jvm_property"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve all property/value pairs
    """
    return isvgAppliance.invoke_get("Retrieve all property/value pair", "{0}".format(uri))


def get(isvgAppliance,  propertyName, check_mode=False, force=False):
    """
    Retrieve specific property/value pair
    """
    return isvgAppliance.invoke_get("Retrieve a specific property/value pair", "{0}?propertyName={1}".format(uri, propertyName))


def search(isvgAppliance, propertyName, check_mode=False, force=False):
    """
    Search for existing property/value pair for a given file. 
    Just care for presence of property, not its actual value.
    """
    ret_obj = get(isvgAppliance, propertyName)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if 'Error' not in obj:
            if obj['propertyName'] == propertyName:
                logger.info("Found property {0}: {1}".format(propertyName, obj['propertyValue']))
                return_obj['data'] = obj
                return_obj['rc'] = 0

    return return_obj


def add(isvgAppliance, propertyName, propertyValue, check_mode=False, force=False):
    """
    Create an new property/value pair
    """
    ret_obj = search(isvgAppliance, propertyName, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the property/value pair
            property_json = {
                "propertyName": propertyName,
                "propertyValue": propertyValue
            }

            return isvgAppliance.invoke_post(
                "Create a new property/value pair", uri, property_json, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def delete(isvgAppliance, propertyName, check_mode=False, force=False):
    """
    Delete a property/value pair
    """
    ret_obj = search(isvgAppliance, propertyName, check_mode=check_mode, force=force)
    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isvgAppliance.invoke_delete(
                "Delete a property/value pair", "{0}?propertyName={1}".format(uri, propertyName), warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def update(isvgAppliance, propertyName, propertyValue, check_mode=False, force=False):
    """
    Update a property/value pair
    """
    ret_obj = search(isvgAppliance, propertyName, check_mode=check_mode, force=force)
    warnings = ret_obj['warnings']

    needs_update = True
    if 'propertyName' in ret_obj['data']:
        if ret_obj['data']['propertyName'] == propertyName and ret_obj['data']['propertyValue'] == propertyValue:
            needs_update = False

    if force is True or needs_update is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the property/value pair
            property_json = {
                "propertyName": propertyName,
                "propertyValue": propertyValue
            }

            return isvgAppliance.invoke_put(
                "Update an existing property/value pair", uri, property_json, warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def set(isvgAppliance, propertyName, propertyValue, check_mode=False, force=False):
    """
    Creating or Modifying a property/value pair
    """
    if (search(isvgAppliance, propertyName))['data'] == {}:
        # Force the add - we already know object does not exist
        logger.info("Property {0} had no match, requesting to add new one.".format(propertyName))
        return add(isvgAppliance, propertyName, propertyValue, check_mode=check_mode, force=True)
    else:
        # Update request
        logger.info("Property {0} exists, requesting to update.".format(propertyName))
        return update(isvgAppliance, propertyName, propertyValue, check_mode=check_mode, force=force)



import logging

logger = logging.getLogger(__name__)

uri = "/mga/user_registry/users"
requires_modules = ["mga", "federation"]
requires_version = None

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all current users in the registry
    """
    return isamAppliance.invoke_get("Retrieving a list of all current users in the registry",
                                    "{0}/v1".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, id, check_mode=False, force=False):
    """
    Retrieving details for a particular user in the registry
    """
    return isamAppliance.invoke_get("Retrieving details for a particular user in the registry",
                                    "{0}/{1}/v1".format(uri, id),
                                    requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, id, password, groups=None, check_mode=False, force=False):
    """
    Creating a new user in the registry
    """
    if force is True or _check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "id": id,
                "password": password
            }
            if groups is not None:
                if isinstance(groups, basestring):
                    import ast
                    groups = ast.literal_eval(groups)
                json_data['groups'] = groups
            return isamAppliance.invoke_post(
                "Creating a new user in the registry",
                "{0}/v1".format(uri), json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Deleting a user in the registry
    """
    if force is True or _check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a user in the registry",
                "{0}/{1}/v1".format(uri, id),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def set_pw(isamAppliance, id, password, check_mode=False, force=False):
    """
    Changing the password of a user in the registry
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Changing the password of a user in the registry",
            "{0}/{1}/v1".format(uri, id),
            {
                'password': password
            },
            requires_modules=requires_modules, requires_version=requires_version)


def _check(isamAppliance, id):
    """
    Check if user already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['id'] == id:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Users between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        user = get(isamAppliance1, obj['id'])
        obj['groups'] = user['data']['groups']
    for obj in ret_obj2['data']:
        user = get(isamAppliance2, obj['id'])
        obj['groups'] = user['data']['groups']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[''])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
am_uri = "/iam/access/v8/attribute-matchers/"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of attribute matchers
    """
    return isamAppliance.invoke_get("Retrieve a list of Attribute Matchers",
                                    "{0}/{1}".format(am_uri, tools.create_query_string(filter=filter, sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, description, check_mode=False, force=False):
    """
    Retrieve a specific attribute matcher
    """
    warnings = []
    ret_obj = search(isamAppliance, description=description, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("Attribute Matcher {0} had no match, skipping retrieval.".format(description))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific Attribute Matcher",
                                    "{0}/{1}".format(am_uri, id))


def search(isamAppliance, description, force=False, check_mode=False):
    """
    Search attribute matcher id by description
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['description'] == description:
            logger.info("Found Attribute Matcher {0} id: {1}".format(description, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, description, properties, predefined=True, supportedDatatype="None", uri="None", check_mode=False,
        force=False):
    """
    Modifying an Attribute Matcher - Add not supported
    """
    warnings = []
    if (search(isamAppliance, description=description))['data'] == {}:
        # Attribute Matcher doesn't exist, Add is not supported
        warnings.append("Attribute Matcher {0} had no match, Add is not supported.".format(description))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        if (description == "Exact attribute matcher" or description == "JavascriptPIPMatcher"):
            warnings.append(
                "Properties for the Exact Attribute Matcher and the JavaScript PIP Matcher should not be modified.")
            return isamAppliance.create_return_object(changed=False, warnings=warnings)
        else:
            # Update request
            logger.info("Attribute Matcher {0} exists, requesting to update.".format(description))
            return update(isamAppliance, description, properties, check_mode, force)


def update(isamAppliance, description, properties, check_mode=False, force=False):
    """
    Update a specified Attribute Matcher
    """
    id, update_required, json_data = _check(isamAppliance, description, properties)
    if id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown Attribute Matcher: {0}".format(description))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified Attribute Matcher",
                "{0}/{1}".format(am_uri, id), json_data)

    return isamAppliance.create_return_object()


def _check(isamAppliance, description, properties):
    """
    Check and return True if update needed
    """
    update_required = False
    ret_obj = get(isamAppliance, description)
    json_data = {}
    if ret_obj['data'] == {}:
        logger.warning("Attribute Matcher not found, returning no update required.")
        return None, update_required, json_data
    else:
        json_data = {
            "properties": properties,
            "predefined": ret_obj['data']['predefined'],
            "supportedDatatype": ret_obj['data']['supportedDatatype'],
            "uri": ret_obj['data']['uri']
        }
        id = ret_obj['data']['id']
        del ret_obj['data']['id']
        # del ret_obj['data']['properties']['id']
        del ret_obj['data']['description']
        count = 0
        for i in ret_obj['data']['properties']:
            del ret_obj['data']['properties'][count]['id']
            count += 1
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Attribute Matchers between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)
    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



import logging
import copy
import json
from ibmsecurity.utilities.tools import version_compare
from ibmsecurity.utilities.tools import json_compare
from ibmsecurity.utilities.tools import jsonSortedListEncoder

logger = logging.getLogger(__name__)

uri = "/isam/cluster/v2"
requires_modules = None
requires_version = None
requires_model = "Appliance"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current cluster configuration
    """
    return isamAppliance.invoke_get("Retrieve the current cluster configuration", uri,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def set(isamAppliance, primary_master='127.0.0.1', secondary_master=None, master_ere=None, tertiary_master=None,
       quaternary_master=None, dsc_external_clients=False, dsc_port=None, dsc_use_ssl=None, dsc_ssl_keyfile=None,
       dsc_ssl_label=None, dsc_worker_threads=64, dsc_maximum_session_lifetime=3600, dsc_client_grace_period=600,
       hvdb_embedded=True, hvdb_max_size=None, hvdb_db_type=None, hvdb_address=None, hvdb_port=None, hvdb_user=None,
       hvdb_password=None, hvdb_db2_alt_address=None, hvdb_db2_alt_port=None, hvdb_db_name=None, hvdb_db_secure=None,
       hvdb_driver_type=None, hvdb_solid_tc=None, cfgdb_embedded=True, cfgdb_db_type=None, cfgdb_address=None,
       cfgdb_port=None, cfgdb_user=None, cfgdb_password=None, cfgdb_db2_alt_address=None, cfgdb_db2_alt_port=None,
       cfgdb_db_name=None, cfgdb_db_secure=None, cfgdb_driver_type=None, cfgdb_solid_tc=None, first_port=2020,
       cfgdb_fs=None, dsc_trace_level=None, cfgdb_db_truststore=None,
       dsc_connection_idle_timeout=None, hvdb_failover_servers=None, cfgdb_failover_servers=None, hvdb_db_truststore=None,
       ignore_password_for_idempotency=False, check_mode=False, force=False):
    """
    Set cluster configuration
    """

    warnings = []
    # Create a simple json with just the main client attributes
    cluster_json = {
        "primary_master": primary_master,
        "dsc_external_clients": bool(dsc_external_clients),
        "dsc_worker_threads": int(dsc_worker_threads),
        "dsc_maximum_session_lifetime": int(dsc_maximum_session_lifetime),
        "dsc_client_grace_period": int(dsc_client_grace_period),
        "hvdb_embedded": bool(hvdb_embedded),
        "cfgdb_embedded": bool(cfgdb_embedded),
        "first_port": int(first_port)
    }
    # Add attributes that have been supplied... otherwise skip them.
    if secondary_master is not None:
        cluster_json["secondary_master"] = secondary_master
    if master_ere is not None:
        cluster_json["master_ere"] = master_ere
    if tertiary_master is not None:
        cluster_json["tertiary_master"] = tertiary_master
    if quaternary_master is not None:
        cluster_json["quaternary_master"] = quaternary_master
    if dsc_port is not None:
        cluster_json["dsc_port"] = int(dsc_port)
    if dsc_use_ssl is not None:
        cluster_json["dsc_use_ssl"] = bool(dsc_use_ssl)
    if dsc_ssl_keyfile is not None:
        cluster_json["dsc_ssl_keyfile"] = dsc_ssl_keyfile
    if dsc_ssl_label is not None:
        cluster_json["dsc_ssl_label"] = dsc_ssl_label
    if dsc_connection_idle_timeout is not None:
        cluster_json["dsc_connection_idle_timeout"] = int(dsc_connection_idle_timeout)
    if hvdb_max_size is not None and hvdb_embedded:
        cluster_json["hvdb_max_size"] = int(hvdb_max_size)
    if hvdb_db_type is not None:
        cluster_json["hvdb_db_type"] = hvdb_db_type
    if hvdb_address is not None:
        cluster_json["hvdb_address"] = hvdb_address
    if hvdb_port is not None:
        cluster_json["hvdb_port"] = int(hvdb_port)
    if hvdb_user is not None:
        cluster_json["hvdb_user"] = hvdb_user
    if hvdb_password is not None:
        if ignore_password_for_idempotency:
            warnings.append("Request made to ignore hvdb_password for idempotency check.")
        else:
            warnings.append("Since existing hvdb_password cannot be read - this call will not be idempotent.")
        cluster_json["hvdb_password"] = hvdb_password
    if hvdb_db2_alt_address is not None:
        cluster_json["hvdb_db2_alt_address"] = hvdb_db2_alt_address
    if hvdb_db2_alt_port is not None:
        cluster_json["hvdb_db2_alt_port"] = int(hvdb_db2_alt_port)
    if hvdb_db_name is not None:
        cluster_json["hvdb_db_name"] = hvdb_db_name
    if hvdb_db_secure is not None:
        cluster_json["hvdb_db_secure"] = bool(hvdb_db_secure)
    if hvdb_driver_type is not None and hvdb_db_type == "oracle":
        #Oracle only
        cluster_json["hvdb_driver_type"] = hvdb_driver_type
    if hvdb_solid_tc is not None:
        if isinstance(hvdb_solid_tc, basestring):
            import ast
            hvdb_solid_tc = ast.literal_eval(hvdb_solid_tc)
        cluster_json["hvdb_solid_tc"] = hvdb_solid_tc
    if hvdb_db_truststore is not None and hvdb_db_truststore != "":
        cluster_json["hvdb_db_truststore"] = hvdb_db_truststore
    if hvdb_failover_servers is not None:
        #Only if the db_type is Postgresql.  This is an array [{"address": "postgresql2.domain.com", "port": 5432, "order": 0}]
        if hvdb_db_type is not None and hvdb_db_type == "postgresql":
            cluster_json["hvdb_failover_servers"] = hvdb_failover_servers
    if cfgdb_failover_servers is not None:
        #Only if the dbtype isPostgresql.  This is an array [{"address": "postgresql2.domain.com", "port": 5432, "order": 0}]
        if cfgdb_db_type is not None and cfgdb_db_type == "postgresql":
            cluster_json["cfgdb_failover_servers"] = cfgdb_failover_servers
    if cfgdb_db_type is not None:
        cluster_json["cfgdb_db_type"] = cfgdb_db_type
    if cfgdb_address is not None:
        cluster_json["cfgdb_address"] = cfgdb_address
    if cfgdb_port is not None:
        cluster_json["cfgdb_port"] = int(cfgdb_port)
    if cfgdb_user is not None:
        cluster_json["cfgdb_user"] = cfgdb_user
    if cfgdb_password is not None:
        if ignore_password_for_idempotency:
            warnings.append("Request made to ignore cfgdb_password for idempotency check.")
        else:
            warnings.append("Since existing cfgdb_password cannot be read - this parameter is ignored for idempotency.")
        cluster_json["cfgdb_password"] = cfgdb_password
    if cfgdb_db2_alt_address is not None:
        cluster_json["cfgdb_db2_alt_address"] = cfgdb_db2_alt_address
    if cfgdb_db2_alt_port is not None:
        cluster_json["cfgdb_db2_alt_port"] = int(cfgdb_db2_alt_port)
    if cfgdb_db_name is not None:
        cluster_json["cfgdb_db_name"] = cfgdb_db_name
    if cfgdb_db_secure is not None:
        cluster_json["cfgdb_db_secure"] = bool(cfgdb_db_secure)
    if cfgdb_driver_type is not None and cfgdb_db_type == "oracle":
        cluster_json["cfgdb_driver_type"] = cfgdb_driver_type
    if cfgdb_solid_tc is not None:
        if isinstance(cfgdb_solid_tc, basestring):
            import ast
            cfgdb_solid_tc = ast.literal_eval(cfgdb_solid_tc)
        cluster_json["cfgdb_solid_tc"] = cfgdb_solid_tc
    if cfgdb_fs is not None:
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.2.0") < 0:
            warnings.append(
                "Appliance at version: {0}, cfgdb_fs: {1} is not supported. Needs 9.0.2.0 or higher. Ignoring cfgdb_fs for this call.".format(
                    isamAppliance.facts["version"], cfgdb_fs))
        else:
            cluster_json["cfgdb_fs"] = cfgdb_fs
    if cfgdb_db_truststore is not None and cfgdb_db_truststore != "":
        cluster_json["cfgdb_db_truststore"] = cfgdb_db_truststore
    if dsc_trace_level is not None:
        cluster_json["dsc_trace_level"] = dsc_trace_level

    check_obj =  _check(isamAppliance, cluster_json, ignore_password_for_idempotency)
    if check_obj['warnings'] != []:
        warnings.append(check_obj['warnings'][0])

    if force is True or check_obj['value'] is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post("Set cluster configuration", uri, cluster_json,
                                             requires_modules=requires_modules, requires_version=requires_version,
                                             requires_model=requires_model,
                                             warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, cluster_json, ignore_password_for_idempotency):
    """
    Check if provided json values match the configuration on appliance

    :param isamAppliance:
    :param cluster_json:
    :return:
    """

    check_obj = {'value': True, 'warnings':""}

    ret_obj = get(isamAppliance)
    check_obj['warnings'] = ret_obj['warnings']

    if ignore_password_for_idempotency:
        temp = copy.deepcopy(
            cluster_json)  # deep copy necessary: otherwise password parameter would be removed from desired config dict 'cluster_json'. Comparison is done
        if 'hvdb_password' in temp:
            del temp["hvdb_password"]
        if 'cfgdb_password' in temp:
            del temp["cfgdb_password"]
        logger.debug("Passwordless JSON to Apply: {0}".format(temp))
    else:
        temp = cluster_json

    #only compare the keys that are in cluster_json, ignore the rest
    #  note that this may yield false comparison results if you remove parameters from your input
    temp_retobj = {}
    for key, value in ret_obj['data'].items():
        if key in temp:
            temp_retobj[key] = value
        else:
            logger.debug("Ignoring {0}".format(key))

    sorted_ret_obj = json.dumps(temp_retobj, skipkeys=True, sort_keys=True)
    sorted_json_data = json.dumps(temp, skipkeys=True, sort_keys=True)

    logger.debug("Sorted Existing Data:{0}".format(sorted_ret_obj))
    logger.debug("Sorted Desired  Data:{0}".format(sorted_json_data))

    if sorted_ret_obj == sorted_json_data:
        logger.debug("JSON provided already is contained in current appliance configuration.")
        check_obj['value'] = True
    else:
        check_obj['value'] = False
    return check_obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cluster configuration between two appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)
requires_modules = ["mga", "federation"]
requires_model = "Appliance"

def get(isamAppliance, check_mode=False, force=False):
    """
    Get runtime tuning settings
    """
    return isamAppliance.invoke_get("Retrieving runtime tuning parameters",
                                    "/mga/runtime_tuning/v1",
                                    requires_modules=requires_modules,requires_model=requires_model)


def set(isamAppliance, option, value, check_mode=False, force=False):

#    ret_obj = get(isamAppliance)
#    for key, val in ret_obj.items():
#        if key == 'warnings' and val != []:
#            if "Docker" in val[0]:
#                return isamAppliance.create_return_object(warnings=ret_obj['warnings'])
    """
    Set a runtime tuning parameter
    """
    warnings = []
    matches, exists = False, False
    if force is False:
        matches, exists, warnings = _check(isamAppliance, option, value)

    if exists is False:
        warnings.append("Tuning Parameter {0} was not found. set() request will attempt anyway.".format(option))

    if force is True or matches is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Setting a runtime tuning parameter",
                "/mga/runtime_tuning/{0}/v1".format(option),
                {
                    'value': value
                }, requires_modules=requires_modules,requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, option, value):
    """
    Check if tuning parameter option exists and matches value
    """
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    matches = False
    exists = False
    try:
        if str(ret_obj['data'][option]) == str(value):
            logger.info("Tuning parameter {0}/{1} already set.".format(option, value))
            matches = True
        else:
            logger.info(
                "Tuning parameter {0} value does not match {1} != {2}.".format(option, ret_obj['data'][option], value))
        exists = True
    except:
        logger.info("Runtime tuning parameter does not exist")

    return matches, exists, warnings


def reset(isamAppliance, option, check_mode=False, force=False):
    """
    Reset a tuning parameter
    """
    warnings = []
    matches, exists = False, False

    if force is False:
        matches, exists, warnings = _check(isamAppliance, option, None)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True,warnings=warnings)
        else:
            return isamAppliance.invoke_delete(
                "Reset a runtime tuning parameter to default value",
                "/mga/runtime_tuning/{0}/v1".format(option),
                requires_modules=requires_modules,requires_model=requires_model)
    elif exists is False and warnings == []:
        warnings.append("Tuning Parameter {0} was not found. Skipping reset() request.".format(option))

    return isamAppliance.create_return_object(warnings=warnings)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare runtime tuning parameters between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    # Ignore differences between endpoints or listening interfaces (compared separately)
    del ret_obj1['data']['endpoints']
    del ret_obj2['data']['endpoints']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['endpoints'])



import logging
from ibmsecurity.utilities.tools import json_sort, json_compare

logger = logging.getLogger(__name__)
uri = "/isam/host_records"
requires_modules = None
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of host IP addresses
    """
    return isamAppliance.invoke_get(
        "Retrieving the list of host IP addresses",
        uri,
        requires_modules=requires_modules,
        requires_version=requires_version,
    )


def get(isamAppliance, host_address, check_mode=False, force=False):
    """
    Retrieving the list of host names associated with a host IP address
    """
    return isamAppliance.invoke_get(
        "Retrieving the list of host names associated with a host IP address",
        "{0}/{1}/hostnames".format(uri, host_address),
        requires_modules=requires_modules,
        requires_version=requires_version,
    )


def set(isamAppliance, addr, hostnames, check_mode=False, force=False):
    """
    Setting a host record (IP address and host name)

    Sample JSON:
    {"addr":"127.0.0.2","hostnames":[{"name":"test1.ibm.com"}, {"name":"test2.ibm.com"}]}
    """
    add_required = False
    delete_required = False
    if force is False:
        # Check if record exists
        if _check(isamAppliance, addr) is True:
            ret_obj = get(isamAppliance, host_address=addr)
            # Check if hostnames match (case sensitive check)
            if json_sort(ret_obj["data"]) != json_sort(hostnames):
                delete_required = True
                add_required = True
        else:
            add_required = True

    if force is True or delete_required is True:
        # No need to check if record exists - force delete
        delete(isamAppliance, host_address=addr, check_mode=check_mode, force=True)

    if force is True or add_required is True:
        # No need to check for add - force add
        return add(
            isamAppliance,
            addr=addr,
            hostnames=hostnames,
            check_mode=check_mode,
            force=True,
        )

    return isamAppliance.create_return_object()


def add(isamAppliance, addr, hostnames, check_mode=False, force=False):
    """
    Creating a host record (IP address and host name)

    Sample JSON:
    {"addr":"127.0.0.2","hostnames":[{"name":"test1.ibm.com"}, {"name":"test2.ibm.com"}]}
    """
    if force is True or _check(isamAppliance, addr) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a host record (IP address and host name)",
                uri,
                {"addr": addr, "hostnames": hostnames},
                requires_modules=requires_modules,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def delete(isamAppliance, host_address, check_mode=False, force=False):
    """
    Removing a host record (IP address and associated host names)
    """
    if force is True or _check(isamAppliance, host_address) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Removing a host record (IP address and associated host names)",
                "{0}/{1}".format(uri, host_address),
                requires_modules=requires_modules,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def _check(isamAppliance, addr, check_mode=False, force=False):
    """
    check whether addr has been defined
    """
    ret_obj = get_all(isamAppliance)

    for hosts in ret_obj["data"]:
        if hosts["addr"] == addr:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Host Records between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/policysets"


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of policy sets
    """
    return isamAppliance.invoke_get(
        "Retrieve a list of policy sets",
        "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)),
    )


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific policy set
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    pol_id = ret_obj["data"]

    if pol_id == {}:
        logger.info("Policy {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve a specific policy set", "{0}/{1}".format(uri, pol_id)
        )


def get_policies(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve policies in a specific policy set
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    pol_id = ret_obj["data"]

    if pol_id == {}:
        logger.info("Policy {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get(
            "Retrieve policies in a specific policy set",
            "{0}/{1}/policies".format(uri, pol_id),
        )


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search policy set id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj["data"]:
        if obj["name"] == name:
            logger.info("Found Policy Set {0} id: {1}".format(name, obj["id"]))
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0

    return return_obj


def set(
    isamAppliance,
    name,
    policies=None,
    policyCombiningAlgorithm="denyOverrides",
    description="",
    predefined=False,
    new_name=None,
    check_mode=False,
    force=False,
):
    """
    Creating or Modifying a Policy Set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    if (search(isamAppliance, name=name))["data"] == {}:
        # Force the add - we already know policy set does not exist
        logger.info(
            "Policy Set {0} had no match, requesting to add new one.".format(name)
        )
        return add(
            isamAppliance,
            name,
            policies,
            policyCombiningAlgorithm,
            description,
            predefined,
            check_mode,
            True,
        )
    else:
        # Update request
        logger.info("Policy Set {0} exists, requesting to update.".format(name))
        return update(
            isamAppliance,
            name,
            policies,
            policyCombiningAlgorithm,
            description,
            predefined,
            new_name,
            check_mode,
            force,
        )


def add(
    isamAppliance,
    name,
    policies=None,
    policyCombiningAlgorithm="denyOverrides",
    description="",
    predefined=False,
    check_mode=False,
    force=False,
):
    """
    Create a new Policy Set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj["data"] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "description": description,
                "predefined": predefined,
            }
            if policies is not None:
                json_data["policies"] = _convert_policy_name_to_id(
                    isamAppliance, policies
                )
            if policyCombiningAlgorithm is not None:
                json_data["policyCombiningAlgorithm"] = policyCombiningAlgorithm
            return isamAppliance.invoke_post("Create a new Policy Set", uri, json_data)

    return isamAppliance.create_return_object()


def _convert_policy_name_to_id(isamAppliance, policies):
    pol_ids = []
    import ibmsecurity.isam.aac.access_control.policies

    for pol_name in policies:
        ret_obj = ibmsecurity.isam.aac.access_control.policies.search(
            isamAppliance, pol_name
        )
        pol_id = ret_obj["data"]
        if pol_id != {}:
            pol_ids.append(pol_id)
            logger.debug("Converting policy {0} to ID: {1}".format(pol_name, pol_id))
        else:
            logger.warning("Unable to find policy {0}, skipping.".format(pol_name))

    return pol_ids


def _convert_policy_id_to_name(isamAppliance, policies):
    pol_names = []
    import ibmsecurity.isam.aac.access_control.policies

    for pol_id in policies:
        ret_obj = ibmsecurity.isam.aac.access_control.policies._get(
            isamAppliance, pol_id
        )
        if ret_obj["data"] != {}:
            pol_name = ret_obj["data"]["name"]
            pol_names.append(pol_name)
            logger.debug("Converting policy {0} to Name: {1}".format(pol_id, pol_name))
        else:
            logger.warning("Unable to find policy {0}, skipping.".format(pol_id))

    return pol_names


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a Policy Set
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    pol_id = ret_obj["data"]

    if pol_id == {}:
        logger.info("Policy Set {0} not found, skipping delete.".format(name))
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a Policy Set", "{0}/{1}".format(uri, pol_id)
            )

    return isamAppliance.create_return_object()


def update(
    isamAppliance,
    name,
    policies=None,
    policyCombiningAlgorithm="denyOverrides",
    description="",
    predefined=False,
    new_name=None,
    check_mode=False,
    force=False,
):
    """
    Update a specified policy set

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    pol_id, update_required, json_data = _check(
        isamAppliance,
        name,
        policies,
        policyCombiningAlgorithm,
        description,
        predefined,
        new_name,
    )
    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError(
            "999", "Cannot update data for unknown policy set: {0}".format(name)
        )

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified policy set",
                "{0}/{1}".format(uri, pol_id),
                json_data,
            )

    return isamAppliance.create_return_object()


def _check(
    isamAppliance,
    name,
    policies,
    policyCombiningAlgorithm,
    description,
    predefined,
    new_name,
):
    """
    Check and return True if update needed
    """
    update_required = False
    json_data = {"name": name, "description": description, "predefined": predefined}
    ret_obj = get(isamAppliance, name)
    if ret_obj["data"] == {}:
        logger.warning("Policy Set not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj["data"]["id"]
        if new_name is not None:
            json_data["name"] = new_name
        else:
            json_data["name"] = name
        if policies is not None:
            json_data["policies"] = _convert_policy_name_to_id(isamAppliance, policies)
        else:
            try:
                del ret_obj["data"]["policies"]
            except Exception:
                pass
        if policyCombiningAlgorithm is not None:
            json_data["policyCombiningAlgorithm"] = policyCombiningAlgorithm
        else:
            try:
                del ret_obj["data"]["policyCombiningAlgorithm"]
            except Exception:
                pass
        del ret_obj["data"]["id"]
        del ret_obj["data"]["userlastmodified"]
        del ret_obj["data"]["lastmodified"]
        del ret_obj["data"]["datecreated"]
        # TODO Check if sorting of policies is a valid for comparison (order maybe important!)
        import ibmsecurity.utilities.tools

        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug("Sorted input: {0}".format(sorted_json_data))
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj["data"])
        logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return pol_id, update_required, json_data


def update_policies(
    isamAppliance, name, policies, action, check_mode=False, force=False
):
    """
    Update a specified policy set's policies (add/remove/set)

    Note: Please input policies as an array of policy names (it will be converted to id's)
    """
    pol_id, update_required, json_data = _check_policies(
        isamAppliance, name, policies, action
    )

    if pol_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError(
            "999", "Cannot update data for unknown policy set: {0}".format(name)
        )

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified policy set",
                "{0}/{1}/policies{2}".format(
                    uri, pol_id, tools.create_query_string(action=action)
                ),
                json_data,
            )

    return isamAppliance.create_return_object()


def _check_policies(isamAppliance, name, policies, action):
    update_required = False
    json_data = {}
    pol_ids = []
    ret_obj = get(isamAppliance, name)
    if ret_obj["data"] == {}:
        logger.warning("Policy Set not found, returning no update required.")
        return None, update_required, json_data
    else:
        pol_id = ret_obj["data"]["id"]
        new_pol_ids = _convert_policy_name_to_id(isamAppliance, policies)
        if action == "set":
            pol_ids = new_pol_ids
        else:
            for new_pol_id in new_pol_ids:
                if new_pol_id in ret_obj["data"]["policies"]:
                    exists = True
                else:
                    exists = False
                if action == "add":
                    if exists:
                        logger.info(
                            "Policy ID {0} already exists skipping.".format(new_pol_id)
                        )
                    else:
                        pol_ids.append(new_pol_ids)
                        logger.info(
                            "Policy ID {0} does not exist, appending to list for additon.".format(
                                new_pol_id
                            )
                        )
                elif action == "delete":
                    if exists:
                        pol_ids.append(new_pol_ids)
                        logger.info(
                            "Policy ID {0} exists, appending to list for deletion.".format(
                                new_pol_id
                            )
                        )
                    else:
                        logger.info("Policy ID {0} does not exist skipping.")
                else:
                    from ibmsecurity.appliance.ibmappliance import IBMError

                    raise IBMError("999", "Unknown action: {0}.".format(action))

        json_data["policies"] = pol_ids
        return pol_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Policy Sets between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        obj["policies"] = _convert_policy_id_to_name(isamAppliance1, obj["policies"])
    for obj in ret_obj2["data"]:
        del obj["id"]
        del obj["userlastmodified"]
        del obj["lastmodified"]
        del obj["datecreated"]
        obj["policies"] = _convert_policy_id_to_name(isamAppliance2, obj["policies"])

    return tools.json_compare(
        ret_obj1,
        ret_obj2,
        deleted_keys=["id", "userlastmodified", "lastmodified", "datecreated"],
    )



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/widgets/notifications"

def get(isvgAppliance, check_mode=False, force=False):
    """
    Get Identity Manager data stores status
    """
    return isvgAppliance.invoke_get("Retrieving notifications status", uri)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

requires_modules = ["wga"]
requires_version = "10.0.6.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all SSH public keys for the default admin user
    """
    return isamAppliance.invoke_get("Retrieving keys", "/admin_cfg/ssh-keys/v1")


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Get specific ssh public key for the default admin user
    """
    return isamAppliance.invoke_get("Retrieving key", f"/admin_cfg/ssh-keys/{uuid}/v1")


def get(isamAppliance, name=None, uuid=None, check_mode=False, force=False):
    """
    Get specific ssh key for admin user

    """
    if uuid is not None and name is None:
        return isamAppliance.invoke_get("Retrieving key", f"/admin_cfg/ssh-keys/{uuid}/v1")
    elif name is not None and uuid is None:
        # Get the key based on the name
        allKeys = get_all(isamAppliance)
        logger.debug(allKeys.get('data'))
        uuids = [d.get('uuid') for d in allKeys.get('data') if d['name'] == name]
        # I expect we'll have 0 or 1 results now
        if uuids:
            uuid = uuids[0]
            return isamAppliance.invoke_get("Retrieving key", f"/admin_cfg/ssh-keys/{uuid}/v1")
        else:
            return None
    else:
        # Input error
        raise IBMError("999", "Cannot get the ssh keys for the administrator. Provide exactly 1 of the parameters uuid or name")


def add(isamAppliance, key, name, check_mode=False, force=False):
    """
    Import ssh public key for the default admin user
    """
    if force or not _check(isamAppliance, name):
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Import ssh key",
                "/admin_cfg/ssh-keys/v1",
                    {
                        'key': key,
                        'name': name
                    })

    return isamAppliance.create_return_object()

create = add
def update(isamAppliance, key, name, check_mode=False, force=False):
    """
    Update an existing key
    """
    # We know we need to do an update.  The only thing we can verify is that the comment is part
    update_required = True
    ret_obj = get(isamAppliance, name)
    if not ret_obj:
        # means we cannot find the name
        logger.debug("Cannot find sshkeys by name")
        update_required = False
    elif key:
        # Extract the comment from the new ssh public key
        key_comment = key.split(' ')[-1]
        logger.debug(f"See if {key_comment} for admin exists in a key")
        if key_comment in ret_obj['data']['fingerprint']:
            update_required = False

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True, warnings=warnings)
    elif update_required:
        delete(isamAppliance, name)
        return add(isamAppliance, key, name)

    return isamAppliance.create_return_object()

def set(isamAppliance, key, name, fingerprint=None, check_mode=False, force=False):
    """
    Create key if it does not exist yet, update if it does.

    The fingerprint is the output format of ssh-keygen:  `ssh-keygen -l -f ~/.ssh/id_rsa.pub`
    It is NOT used to send to ISVA, it is used to check if a key already exists.
    """
    if not _check(isamAppliance, name, fingerprint):
        # Force the add - we already know there is no key by this name or fingerprint
        return add(isamAppliance, key=key, name=name,
                   check_mode=check_mode, force=True)
    else:
        # There is no update.  Let's see what this does.
        return update(isamAppliance, key=key, name=name, check_mode=check_mode, force=force)


def _check(isamAppliance, name=None, fingerprint=None):
    """
    The fingerprint is generated from the actual public key, using `ssh-keygen -l -f ~/.ssh/id_rsa.pub`
    I'm unsure how to do this in Python, but I'm adding it to the check anyway (because Ansible)

    If a fingerprint is supplied, that is the only parameter that will be used.
    The only way to have full idempotency, is by having the fingerprint.

    The other checks will not be completely idempotent.

    :param isamAppliance:
    :param name:
    :return:
    """
    ret_obj = get_all(isamAppliance)
    # Fingerprint rules
    if fingerprint:
        for sshkeys in ret_obj['data']:
            if sshkeys.get('fingerprint') == fingerprint:
                logger.debug(f"Fingerprint matched on {sshkeys.get('name')} for the admin user")
                return True

    if name:
        for sshkeys in ret_obj['data']:
            if sshkeys.get('name') == name:
                logger.debug(f"Name found on {sshkeys.get('name')} for the admin user")
                return True
    return False


def delete(isamAppliance, name=None, uuid=None, check_mode=False, force=False):
    """
    Delete a public ssh_key for the default admin user
    """
    ret_obj = get(isamAppliance, name, uuid)
    if force or ret_obj.get('data'):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if not uuid:
                uuid = ret_obj.get('data').get('uuid')
            return isamAppliance.invoke_delete("Deleting key",
                                               f"/admin_cfg/ssh-keys/{uuid}/v1")

    return isamAppliance.create_return_object()

def compare(isamAppliance1, isamAppliance2):
    """
    Compare the list of SSH keys between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all_app(isvgAppliance, check_mode=False, force=False):
    """
    Retrieving all application interfaces
    :rtype: (str, dict)
    """
    return isvgAppliance.invoke_get("Retrieving all application interfaces", "/application_interfaces")


def add(isvgAppliance, address, netmask, network_type, ipFqdn, gateway, interface="P.1", check_mode=False,
        force=False):
    """
    Adding an address to an application interface
    """
    if force is True or _check(isvgAppliance, address, netmask, network_type, ipFqdn, gateway, interface) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post(
                "Configuring address for application interface",
                "/application_interfaces/" + interface + "/addresses",
                {
                    'address': address,
                    'netmask': netmask,
                    'type': network_type,
                    'ipFqdn': ipFqdn,
                    'gateway': gateway
                })

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, address, netmask, network_type, ipFqdn, gateway, interface, check_mode=False,
        force=False):
    """
    Check whether application interface is configured
    """
    check_value = False

    ret_obj = get_all_app(isvgAppliance)

    for intf in ret_obj['data']:
        if intf['id'] == interface and intf['state'] == "up":
            logger.warning("Application interface already configured")
            check_value = True
            break

    return check_value


import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, reverseproxy_id, check_mode=False, force=False):
    """
    Retrieving a list of stanzas - Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieving a list of stanzas - Reverse Proxy",
                                    "/wga/reverseproxy/{0}/configuration/stanza".format(reverseproxy_id))


def add(isamAppliance, reverseproxy_id, stanza_id, check_mode=False, force=False):
    """
    Adding a configuration stanza name - Reverse Proxy
    """
    if force is True or _check(isamAppliance, reverseproxy_id, stanza_id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Adding a configuration stanza name - Reverse Proxy",
                "/wga/reverseproxy/{0}/configuration/stanza/{1}".format(reverseproxy_id, stanza_id),
                {})

    return isamAppliance.create_return_object()


def delete(isamAppliance, reverseproxy_id, stanza_id, check_mode=False, force=False):
    """
    Deleting a stanza - Reverse Proxy
    """
    if force is True or _check(isamAppliance, reverseproxy_id, stanza_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a stanza - Reverse Proxy",
                "/wga/reverseproxy/{0}/configuration/stanza/{1}".format(reverseproxy_id, stanza_id))

    return isamAppliance.create_return_object()


def _check(isamAppliance, reverseproxy_id, stanza_id):
    """
    Check if entry exists
    """
    ret_obj = get(isamAppliance, reverseproxy_id)

    for stanza in ret_obj['data']:
        if stanza == stanza_id:
            logger.info("Stanza found in resource: " + reverseproxy_id)
            return True

    logger.info("Stanza *not* found in resource: " + reverseproxy_id)
    return False


def compare(isamAppliance1, isamAppliance2, reverseproxy_id, reverseproxy_id2=None):
    """
    Compare stanzas within reverse proxy configuration between two appliances
    """
    if reverseproxy_id2 is None or reverseproxy_id2 == '':
        reverseproxy_id2 = reverseproxy_id

    import ibmsecurity.isam.web.reverse_proxy.configuration.entry

    # The following array contains entries that will be ignored across all stanzas
    ignore_entries = ['azn-server-name', 'pd-user-pwd', 'bind-pwd', 'network-interface', 'server-name',
                      'listen-interface']

    # Retrieve all stanzas and corresponding entries for comparison
    ret_obj1 = get(isamAppliance1, reverseproxy_id)
    entries = {}
    for stanza in ret_obj1['data']:
        entries[stanza] = {}
        stanza_entries = ibmsecurity.isam.web.reverse_proxy.configuration.entry.get_all(isamAppliance1, reverseproxy_id,
                                                                                        stanza)
        for k, v in stanza_entries['data'].items():
            if k not in ignore_entries:
                entries[stanza][str(k)] = v
    ret_obj1['data'] = entries

    ret_obj2 = get(isamAppliance2, reverseproxy_id2)
    entries = {}
    for stanza in ret_obj2['data']:
        entries[stanza] = {}
        stanza_entries = ibmsecurity.isam.web.reverse_proxy.configuration.entry.get_all(isamAppliance2, reverseproxy_id,
                                                                                        stanza)
        for k, v in stanza_entries['data'].items():
            if k not in ignore_entries:
                entries[stanza][str(k)] = v
    ret_obj2['data'] = entries

    return ibmsecurity.utilities.tools.json_compare(ret_obj1=ret_obj1, ret_obj2=ret_obj2, deleted_keys=ignore_entries)



import logging

logger = logging.getLogger(__name__)

uri = "/isam/container_ext/image"

requires_version = "10.0.7.0"
requires_model = "Appliance"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving all known images

    """
    return isamAppliance.invoke_get(
        "Retrieving images",
        uri,
        requires_model=requires_model,
        requires_version=requires_version,
    )


def get(isamAppliance, image_id, check_mode=False, force=False):
    """
    Get image properties
    """
    return isamAppliance.invoke_get(
        "Retrieving configuration for image",
        f"{uri}/{image_id}",
        requires_model=requires_model,
        requires_version=requires_version,
    )


def search(isamAppliance, image, check_mode=False, force=False):
    """
    Return the id of the image

    :param isamAppliance:
    :param image:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = get_all(isamAppliance, check_mode, force)
    return_obj = isamAppliance.create_return_object()
    return_obj["data"] = None
    for obj in ret_obj["data"]:
        if obj["image"] == image:
            return_obj["data"] = obj["id"]
            return_obj["rc"] = 0
            logger.debug(f"Found id: {obj['id']} for {image}")
            break

    return return_obj


def add(isamAppliance, image, check_mode=False, force=False, warnings=None):
    """
    Pull an image
    Set force to True to force pull
    OR use update
    """
    if force or not _check(isamAppliance, image):
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            post_data = {"image": image}

            return isamAppliance.invoke_post(
                "Pull image",
                uri,
                post_data,
                requires_model=requires_model,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def set(isamAppliance, image, check_mode=False, force=False, warnings=None):
    if _check(isamAppliance, image):
        return update(isamAppliance, image, check_mode, force, warnings)
    else:
        return add(isamAppliance, image, check_mode, force, warnings)


def update(isamAppliance, image, check_mode=False, force=False, warnings=None):
    """
    Pull a new version of (existing) image
    Only do this if force = true
    """
    image_id = search(isamAppliance, image)
    image_id = image_id.get("data", None)

    if force and image_id is not None:
        logger.debug(f"(Force) Updating {image_id}")
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            put_data = {"image": image}
            return isamAppliance.invoke_put(
                "Force pull image",
                f"{uri}/{image_id}",
                put_data,
                requires_model=requires_model,
                requires_version=requires_version,
            )
    else:
        if warnings is None:
            warnings = [f"Image {image} does not exist yet, or not set to force=True"]
        else:
            warnings.append(f"Image {image} does not exist yet, or not set to force=True")
        return isamAppliance.create_return_object(changed=False, warnings=warnings)


def _check(isamAppliance, name):
    """
    Check if there's a image going by name
    """
    ret_obj = get_all(isamAppliance)
    for c in ret_obj["data"]:
        if c.get("image") == name:
            logger.debug(f"Volume {name} exists")
            return True
    return False


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an image by name
    ./testisam_cmd.py --hostname=${ISAM_LMI} --method=ibmsecurity.isam.base.container_ext.image.delete --method_options="name=icr.io/isva/verify-access-oidc-provider:23.03"
    """
    ret_obj = search(isamAppliance, name)
    image_id = ret_obj.get("data", None)

    if force or image_id is not None:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an image",
                f"{uri}/{image_id}"
            )

    return isamAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)


def get_network(isdsAppliance, application_interface, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Application Interface Statistics
    """
    return isdsAppliance.invoke_get("Retrieving the Application Interface Statistics",
                                    "/analysis/interface_statistics{0}".format(
                                        tools.create_query_string(prefix=application_interface,
                                                                  timespan=statistics_duration)))


def get_cpu(isdsAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the CPU Usage Statistics
    """
    return isdsAppliance.invoke_get(
        "Retrieving the CPU Usage Statistics",
        "/analysis/system_cpu{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)))


def get_memory(isdsAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isdsAppliance.invoke_get(
        "Retrieving the Memory Usage Statistics",
        "/analysis/system_memory{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)))


def get_storage(isdsAppliance, statistics_duration, check_mode=False, force=False):
    """
    Retrieving the Storage Usage Statistics
    """
    return isdsAppliance.invoke_get(
        "Retrieving the Storage Usage Statistics",
        "/analysis/system_storage{0}".format(
            tools.create_query_string(
                timespan=statistics_duration)))



import logging
import ibmsecurity

import ibmsecurity.isam.web.reverse_proxy.instance

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/redis_config/wrp"
requires_modules = ["wga"]
requires_version = "10.0.1"


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve the Redis collections configured for a specific Web Reverse Proxy
    """
    return isamAppliance.invoke_get("Retrieve the Redis collections configured for a specific Web",
                                    "{0}/{1}".format(uri, name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def update(isamAppliance, name, input_data=[], check_mode=False, force=False):
    """
    Update the Redis collections which are used by a Web Reverse Proxy
    """

    exist, warnings = _check_exist(isamAppliance, name)
    if exist is True:
        same_contents, warnings = _check_contents(isamAppliance=isamAppliance, name=name, input_data=input_data)
    else:
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or same_contents is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update the Redis collections which are used by a Web Reverse Proxy",
                "{0}/{1}".format(uri, name),
                input_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_exist(isamAppliance, name):
    ret_obj = ibmsecurity.isam.web.reverse_proxy.instance.get(isamAppliance)

    for rp in ret_obj['data']:
        if rp['id'] == name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_contents(isamAppliance, name, input_data):
    ret_obj = get(isamAppliance, name)

    if ret_obj['data'] == input_data:
        return True, ret_obj['warnings']
    else:
        return False, ret_obj['warnings']


def compare(isamAppliance1, isamAppliance2, name1, name2=None):
    if name2 is None or name1 == '':
        name2 = name1
    ret_obj1 = get(isamAppliance1, name1)
    ret_obj2 = get(isamAppliance2, name2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging
import json
import tempfile
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/custfile_mgmt"


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get keystore info
    """
    return isvgAppliance.invoke_get("Retrieving keystore entries", uri + "/data")


def search(isvgAppliance, keystore="itimKeystore.jceks", check_mode=False, force=False):
    """
    Search for existing library by file name
    """
    ret_obj = get(isvgAppliance)

    return_obj = isvgAppliance.create_return_object()

    if 'children' in ret_obj['data']:
        for child in ret_obj['data']['children']:
            if 'fileName' in child and child['fileName'] == 'keystore':
                files=json.loads(child['files'])
                for file in files:
                    if 'fileName' in file and file['fileName'] == keystore:
                        logger.info("Found keystore entry: {0}".format(file['fileName']))
                        return_obj['data'] = file
                        return_obj['rc'] = 0
                        break

    return return_obj
    

def upload(isvgAppliance, keystore, check_mode=False, force=False):
    """
    Import keystore
    """
    warnings = []

    if force is True or not _check(isvgAppliance, keystore):
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            ret_obj = isvgAppliance.invoke_post_files(
                "Import keystore",
                "/upload_object",
                [
                    {
                        'file_formfield': 'uploadedfile',
                        'filename': keystore,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'fileName': tools.path_leaf(keystore),
                    'upload_file_type': 'custfile'
                },
                json_response=False)

            warnings = ret_obj['warnings']

            if ret_obj["rc"] == 0:
                keystore_json = {
                   "fileName": "itimKeystore.jceks",
                   "filePath": "data/keystore",
                   "ObjType": "externalfile"
                }

                return isvgAppliance.invoke_post(
                    "Save keystore metadata", uri, keystore_json, warnings=warnings)

    return isvgAppliance.create_return_object()


def download(isvgAppliance, filename, check_mode=False, force=False):
    """
    Download keystore
    """
    import os.path

    if force is True or (os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isvgAppliance.invoke_get_file("Downloading IM keystore",
                                                 "{0}/{1}?data/keystore/{2}".format(uri, "download", "itimKeystore.jceks"), filename=filename, mime_types="*/*")

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, keystore):
    """
    Check if keystore needs to be uploaded
    Need to download current keystore in order to compare both
    """
    ret_obj = search(isvgAppliance, tools.path_leaf(keystore))

    if len(ret_obj['data']) > 0:
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpFile = tmpdirname + "/" + tools.path_leaf(keystore)
            ret_obj = download(isvgAppliance, tmpFile)
            if ret_obj["rc"] == 0:
                return tools.files_same(keystore, tmpFile)
    else:
        return False



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all rsyslog objects
    """
    return isamAppliance.invoke_get("Get all rsyslog objects",
                                    "/core/rsp_rsyslog_objs")


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Get a specific rsyslog object
    """
    return isamAppliance.invoke_get("Get a specific rsyslog object",
                                    "/core/rsp_rsyslog_objs/{0}".format(uuid))


def add(isamAppliance, name, collector, collectorPort=514, collectorLeef=False, objType='rsyslog', comment='',
        check_mode=False, force=False):
    """
    Add a rsyslog object
    """
    if force is True or _check(isamAppliance, None, name, objType, comment, collector, collectorPort,
                               collectorLeef) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Add a rsyslog object",
                "/core/rsp_rsyslog_objs/",
                {
                    'name': name,
                    'objType': objType,
                    'comment': comment,
                    'collector': collector,
                    'collectorPort': collectorPort,
                    'collectorLeef': collectorLeef
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, uuid, name, collector, collectorPort=514, collectorLeef=False, objType='rsyslog', comment='',
           check_mode=False, force=False):
    """
    Update a specific rsyslog object
    """
    if force is True or (
            _exists(isamAppliance, uuid) is True and _check(isamAppliance, uuid, name, objType, comment,
                                                            collector, collectorPort, collectorLeef) is False):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific rsyslog object",
                "/core/rsp_rsyslog_objs/{0}".format(uuid),
                {
                    'name': name,
                    'uuid': uuid,
                    'objType': objType,
                    'comment': comment,
                    'collector': collector,
                    'collectorPort': collectorPort,
                    'collectorLeef': collectorLeef
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, uuid, check_mode=False, force=False):
    """
    Delete a rsyslog object
    """
    if force is True or _exists(isamAppliance, uuid) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a rsyslog object",
                "/core/rsp_rsyslog_objs/{0}".format(uuid))

    return isamAppliance.create_return_object()


def _exists(isamAppliance, uuid):
    """
    Check if an uuid object exists

    :param isamAppliance:
    :param uuid:
    :return:
    """
    exists = False
    ret_obj = get_all(isamAppliance)

    for rsyslog in ret_obj['data']['rsyslogObjects']:
        if rsyslog['uuid'] == uuid:
            exists = True
            break

    return exists


def _check(isamAppliance, uuid, name, objType, comment, collector, collectorPort, collectorLeef):
    """
    Check if the rsyslog object exists and is the same - uuid=None means add versus delete

    NOTE: if UUID is not found that will be same as no match!!!
    """

    set_value = {
        'name': name,
        'uuid': uuid,
        'objType': objType,
        'comment': comment,
        'collector': collector,
        'collectorPort': collectorPort,
        'collectorLeef': collectorLeef
    }

    set_value = ibmsecurity.utilities.tools.json_sort(set_value)

    ret_obj = get_all(isamAppliance)
    for obj in ret_obj['data']['rsyslogObjects']:
        if uuid is None and obj['name'] == name:
            return True
        elif ibmsecurity.utilities.tools.json_sort(obj) == set_value:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare rsyslog objects between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']['rsyslogObjects']:
        del obj['uuid']
    for obj in ret_obj2['data']['rsyslogObjects']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import logging
import os.path
import ibmsecurity.utilities.tools
from io import open

logger = logging.getLogger(__name__)


def get_all(isamAppliance, kdb_id, check_mode=False, force=False):
    """
    Retrieving personal certificate names and details in a certificate database
    """
    return isamAppliance.invoke_get(
        "Retrieving personal certificate names and details in a certificate database",
        "/isam/ssl_certificates/{0}/personal_cert".format(kdb_id)
    )


def get(isamAppliance, kdb_id, cert_id, check_mode=False, force=False):
    """
    Retrieving a personal certificate from a certificate database
    """
    return isamAppliance.invoke_get(
        "Retrieving a personal certificate from a certificate database",
        "/isam/ssl_certificates/{0}/personal_cert/{1}".format(kdb_id, cert_id)
    )


def generate(isamAppliance, kdb_id, label, dn, expire='365', default='no', size='2048', signature_algorithm='',
             check_mode=False, force=False):
    """
    Generating a self-signed personal certificate in a certificate database
    """
    warnings = []

    if signature_algorithm is not None:
        if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "9.0.2.0") < 0:
            warnings.append("Appliance at version: {0}, signature_algorithm is not supported. Needs 9.0.2.0 or higher. "
                            "Ignoring signature_algorithm for this call".format(isamAppliance.facts["version"]))
            json_obj = {
                "operation": 'generate',
                "label": label,
                "dn": dn,
                "expire": expire,
                'default': default,
                'size': size
            }
        else:
            json_obj = {
                "operation": 'generate',
                "label": label,
                "dn": dn,
                "expire": expire,
                'default': default,
                'size': size,
                'signature_algorithm': signature_algorithm
            }

    if force is True or _check(isamAppliance, kdb_id, label) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Generating a self-signed personal certificate in a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert".format(kdb_id),
                json_obj,
                warnings=warnings
            )

    return isamAppliance.create_return_object(changed=False, warnings=warnings)


def rename(isamAppliance, kdb_id, cert_id, new_id,
           check_mode=False, force=False):
    """
    Rename a personal certificate.  New in 10.0.7
    """
    warnings = []

    if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.7.0") < 0:
        warnings.append(
            f"Appliance is at version: {isamAppliance.facts['version']}. Renaming a certificate requires at least 10.0.7.0"
        )
    else:
        if force or (_check(isamAppliance, kdb_id, cert_id) and not _check(isamAppliance, kdb_id, new_id)):
            if check_mode:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Renaming a personal certificate as default in a certificate database",
                    f"/isam/ssl_certificates/{kdb_id}/personal_cert/{cert_id}",
                    {
                        'new_id': new_id
                    })


def set(isamAppliance, kdb_id, cert_id, default='no', check_mode=False, force=False):
    """
    Setting a personal certificate as default in a certificate database

    Obsolete since 10.0.3
    """
    warnings = []

    if ibmsecurity.utilities.tools.version_compare(isamAppliance.facts["version"], "10.0.3.0") > 0:
        warnings.append(
            f"Appliance is at version: {isamAppliance.facts['version']}. Setting certificates as default is no longer supported."
        )
    else:
        if force is True or _check_default(isamAppliance, kdb_id, cert_id, default):
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Setting a personal certificate as default in a certificate database",
                    "/isam/ssl_certificates/{0}/personal_cert/{1}".format(kdb_id, cert_id),
                    {
                        'default': default
                    })

    return isamAppliance.create_return_object(warnings=warnings)


def receive(isamAppliance, kdb_id, label, cert, default='no', check_mode=False, force=False):
    """
    Receiving a personal certificate into a certificate database
    """
    if force is True or _check(isamAppliance, kdb_id, label) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Receiving a personal certificate into a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert".format(kdb_id),
                [
                    {
                        'file_formfield': 'cert',
                        'filename': cert,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'default': default,
                    'operation': 'receive'
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, kdb_id, cert_id, check_mode=False, force=False):
    """
    Deleting a personal certificate from a certificate database
    """
    if force is True or _check(isamAppliance, kdb_id, cert_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a personal certificate from a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert/{1}".format(kdb_id, cert_id))

    return isamAppliance.create_return_object()


def export_cert(isamAppliance, kdb_id, cert_id, filename, check_mode=False, force=False):
    """
    Exporting a personal certificate from a certificate database
    """
    if force is True or (_check(isamAppliance, kdb_id, cert_id) is True and os.path.exists(filename) is False):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Exporting a personal certificate from a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert/{1}?export".format(kdb_id, cert_id),
                filename)

    return isamAppliance.create_return_object()


def import_cert(isamAppliance, kdb_id, cert, label=None, password=None, check_mode=False, force=False):
    """
    Importing a personal certificate into a certificate database
    Remark: you can add a label, but it's only used for a half-hearted check if the certificate already exists.
    """
    if force is True or not _check(isamAppliance, kdb_id, label):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a personal certificate into a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert".format(kdb_id),
                [
                    {
                        'file_formfield': 'cert',
                        'filename': cert,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'password': password,
                    'operation': 'import'
                }
            )

    return isamAppliance.create_return_object()


def extract_cert(isamAppliance, kdb_id, cert_id, password, filename, check_mode=False, force=False):
    """
    Extracting a personal certificate from a certificate database
    """
    if force is True or (_check(isamAppliance, kdb_id, cert_id) is True and os.path.exists(filename) is False):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = isamAppliance.invoke_post(
                "Extracting a personal certificate from a certificate database",
                "/isam/ssl_certificates/{0}/personal_cert/{1}".format(kdb_id, cert_id),
                {
                    'operation': 'extract',
                    'type': 'pkcs12',
                    'password': password
                })
            extracted_file = open(filename, 'wb')
            extracted_file.write(ret_obj['data'])
            extracted_file.close()

    return isamAppliance.create_return_object()


def _check(isamAppliance, kdb_id, cert_id):
    """
    Check if personal certificate already exists in certificate database
    """
    if cert_id is None:
        logger.debug("No label passed, so return false")
        return False
    else:
        ret_obj = get_all(isamAppliance, kdb_id)
        for certdb in ret_obj['data']:
            if certdb['id'] == cert_id:
                return True

    return False


def _check_default(isamAppliance, kdb_id, cert_id, default):
    """
    Check if personal certificate already exists in certificate database
    """
    ret_obj = get_all(isamAppliance, kdb_id)

    for certdb in ret_obj['data']:
        if certdb['id'] == cert_id:
            if (certdb['default'].lower() == 'true' and default.lower() == 'no') or (
                    certdb['default'].lower() == 'false' and default.lower() == 'yes'):
                return True

    return False


def compare(isamAppliance1, isamAppliance2, kdb_id):
    """
    Compare signer certificates in certificate database between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, kdb_id)
    ret_obj2 = get_all(isamAppliance2, kdb_id)

    for cert in ret_obj1['data']:
        del cert['issuer']
        del cert['notafter']
        del cert['notafter_epoch']
        del cert['notbefore']
        del cert['notbefore_epoch']
        del cert['serial_number']
        del cert['sha1_fingerprint']
        del cert['subject']
    for cert in ret_obj2['data']:
        del cert['issuer']
        del cert['notafter']
        del cert['notafter_epoch']
        del cert['notbefore']
        del cert['notbefore_epoch']
        del cert['serial_number']
        del cert['sha1_fingerprint']
        del cert['subject']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2,
                                                    deleted_keys=['issuer', 'notafter', 'notafter_epoch', 'notbefore',
                                                                  'notbefore_epoch', 'serial_number',
                                                                  'sha1_fingerprint', 'subject'])



import logging

logger = logging.getLogger(__name__)
uri = "/isam/host_records"
requires_modules = None
requires_version = None


def add(isamAppliance, host_address, hostname, check_mode=False, force=False):
    """
    Adding a host name to a host IP address
    """
    if force or not _check(isamAppliance, host_address, hostname):
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Adding a host name to a host IP address",
                "{0}/{1}/hostnames".format(uri, host_address),
                {"name": hostname},
                requires_modules=requires_modules,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def delete(isamAppliance, host_address, hostname, check_mode=False, force=False):
    """
    Removing a host name from a host IP address
    """
    if force or _check(isamAppliance, host_address, hostname):
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Removing a host name from a host IP address",
                "{0}/{1}/hostnames/{2}".format(uri, host_address, hostname),
                requires_modules=requires_modules,
                requires_version=requires_version,
            )

    return isamAppliance.create_return_object()


def _check(isamAppliance, host_address, hostname):
    """
    Check whether existing host record contains a host name
    """
    import ibmsecurity.isam.base.host.records

    try:
        ret_obj = ibmsecurity.isam.base.host.records.get(
            isamAppliance, host_address=host_address
        )
    except Exception:
        from ibmsecurity.appliance.ibmappliance import IBMError

        raise IBMError(
            "999", "Unable to find host record with adress: {0}".format(host_address)
        )

    for hosts in ret_obj["data"]:
        if hosts["name"].lower() == str(hostname).lower():
            return True

    return False



import logging

logger = logging.getLogger(__name__)


def set_pw(isamAppliance, password, check_mode=False, force=False):
    """
    Changing the administrator password of the embedded LDAP server
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post("Changing the administrator password of the embedded LDAP server",
                                         "/isam/embedded_ldap/change_pwd/v1",
                                         {
                                             "password": password
                                         })



import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)
uri = "/isam/dsc/admin/replicas"
requires_modules = None
requires_version = None
requires_model = "Appliance"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of replica sets
    """
    return isamAppliance.invoke_get("Retrieving the list of replica sets",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def get_servers(isamAppliance, replica, check_mode=False, force=False):
    """
    Retrieving the list of servers for a replica set
    """
    return isamAppliance.invoke_get("Retrieving the list of servers for a replica set",
                                    "{0}/{1}/servers".format(uri, replica),
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def get_session(isamAppliance, replica, user="*", max="1024", check_mode=False, force=False):
    """
    Searching for session within a replica set
    """
    return isamAppliance.invoke_get("Searching for session within a replica set",
                                    "{0}/{1}/sessions{2}".format(uri, replica,
                                                                 tools.create_query_string(user=user, max=max)),
                                    requires_modules=requires_modules, requires_version=requires_version,requires_model=requires_model)


def delete(isamAppliance, replica, session, check_mode=False, force=False):
    """
    Terminating a session
    """
    check_value, warnings = _check(isamAppliance)

    if check_value is True:
        if force is True or search(isamAppliance=isamAppliance, replica=replica, session=session) is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return isamAppliance.invoke_delete("Terminating a session",
                                                   "{0}/{1}/sessions/session/{2}".format(uri, replica, session),
                                                   requires_modules=requires_modules, requires_version=requires_version,
                                                   requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def delete_all(isamAppliance, replica, username, check_mode=False, force=False):
    """
    Terminating all sessions for a single user
    """
    check_value, warnings = _check(isamAppliance)

    if check_value is True:
        if force is True or search(isamAppliance=isamAppliance, replica=replica, username=username) is True:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True, warnings=warnings)
            else:
                return isamAppliance.invoke_delete("Terminating all sessions for a single user",
                                                   "{0}/{1}/sessions/user/{2}".format(uri, replica, username),
                                                   requires_modules=requires_modules,
                                                   requires_version=requires_version, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def search(isamAppliance, replica, session=None, username=None, check_mode=False, force=False):

    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']

    if warnings and 'Docker' in warnings[0]:
        return False

    if session != None:
        ret_obj = get_session(isamAppliance, replica)
        data = ret_obj['data']['matched_sessions']
        for obj in data:
            if obj['session'] == session:
                return True
    elif username != None:
        ret_obj = get_session(isamAppliance, replica, user=username, max=1)
        data = ret_obj['data']['matched_sessions']
        if len(data) == 1:
            return True


def _check(isamAppliance):
    """
    Check if it's appliance or not
    :param isamAppliance:
    :return: true|false, warnings message
    """
    ret_obj = get_all(isamAppliance)
    check_value, warnings=False, ret_obj['warnings']

    if warnings == []:
        check_value = True
        return check_value, warnings
    else:
        return check_value, warnings



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isdsAppliance, check_mode=False, force=False):
    """
    Get advanced tuning parameters
    """
    return isdsAppliance.invoke_get("Get advanced tuning parameters",
                                    "/adv_params")


def set(isdsAppliance, key, value, comment="", check_mode=False, force=False):
    """
    Set advanced tuning parameter

    Note: Pass an array of values if needed to set a set of values for given key

    Advanced Tuning Parameters functionality available starting with 8.0.1.9 only
    """

    if ibmsecurity.utilities.tools.version_compare(isdsAppliance.facts["version"], "8.0.1.9") < 0:
        warnings = []
        warnings.append(
            "Appliance at version: {0}, 'Advanced Tuning Parameters' is not supported. Needs 8.0.1.9 or higher. Ignoring 'Advanced Tuning Parameters' for this call.".format(
                isdsAppliance.facts["version"]))
        return isdsAppliance.create_return_object(warnings=warnings)

    ret_obj = isdsAppliance.create_return_object()
    ret_obj['data'] = []
    rc, uuid = False, []
    # Force the value to be an array if not already so
    if not isinstance(value, list):
        value = [value]
    (rc, uuid) = _check(isdsAppliance, key, value)

    # See if change detected or are being forced to do so
    if rc is False or force is True:
        if check_mode is True:
            ret_obj['changed'] = True
        else:
            # Remove any existing values...
            if len(uuid) > 0:
                _del(isdsAppliance, uuid)
            # Add all new/modified values
            for v in value:
                r = isdsAppliance.invoke_post(
                    "Adding advanced tuning parameter",
                    "/adv_params",
                    {
                        '_isNew': True,
                        'comment': comment,
                        'key': key,
                        'value': v
                    })
                ret_obj['changed'] = ret_obj['changed'] or r['changed']
                ret_obj['data'].append(r['data'])
                ret_obj['warnings'].extend(r['warnings'])
                ret_obj['rc'] += r['rc']

    return ret_obj


def _check(isdsAppliance, key, value=None):
    ret_obj = get(isdsAppliance)

    rc = False
    uuid = []
    exist_value_list = []
    for param in ret_obj['data']['tuningParameters']:
        if param['key'] == key:
            exist_value_list.append(str(param['value']))
            uuid.append(param['uuid'])
            logger.info("Advanced tuning parameter key: {0} and value: {1} exists".format(key, value))

    # value being none - means called from delete function
    if value is not None:
        given_value_list = []
        for v in value:
            given_value_list.append(str(v))
        rc = ibmsecurity.utilities.tools.json_sort(exist_value_list) == ibmsecurity.utilities.tools.json_sort(
            given_value_list)
        logger.info(
            "Advanced tuning parameter: {0} has existing values: {1}, and given values: {2} - match status: {3}".format(
                key, exist_value_list, given_value_list, rc))

    return rc, uuid


def delete(isdsAppliance, key, check_mode=False, force=False):
    """
    Delete advanced tuning parameter
    """
    ret_obj = isdsAppliance.create_return_object()
    (rc, uuid) = _check(isdsAppliance, key)

    if len(uuid) > 0:
        if check_mode is True:
            ret_obj['changed'] = True
        else:
            ret_obj = _del(isdsAppliance, uuid)

    return ret_obj


def _del(isdsAppliance, uuid):
    """
    Remove all UUIDs - should be everything pertaining to an advanced tuning parameter "key"

    :param isdsAppliance:
    :param uuid:
    :return:
    """
    ret_obj = isdsAppliance.create_return_object()
    ret_obj['data'] = []

    for u in uuid:
        r = isdsAppliance.invoke_delete(
            "Deleting existing advanced tuning parameter",
            "/adv_params/{0}".format(u))
        ret_obj['changed'] = ret_obj['changed'] or r['changed']
        ret_obj['data'].append(r['data'])
        ret_obj['warnings'].extend(r['warnings'])
        ret_obj['rc'] += r['rc']

    return ret_obj


def compare(isdsAppliance1, isdsAppliance2):
    """
    Compare advanced tuning parameters between two appliances
    """
    ret_obj1 = get(isdsAppliance1)
    ret_obj2 = get(isdsAppliance2)

    # Ignore differences between comments/uuid as they are immaterial
    for param in ret_obj1['data']['tuningParameters']:
        del param['comment']
        del param['uuid']
    for param in ret_obj2['data']['tuningParameters']:
        del param['comment']
        del param['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['comment', 'uuid'])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/resource/instance"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of all Reverse Proxy Instances
    """
    return isamAppliance.invoke_get("Retrieve a list of all Reverse Proxy Instances",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, instance_name, check_mode=False, force=False):
    """
    Retrieve a single Reverse Proxy Instance
    """
    return isamAppliance.invoke_get("Retrieve a single Reverse Proxy Instance",
                                    "{0}/{1}".format(uri, instance_name),
                                    requires_modules=requires_modules, requires_version=requires_version)


def compare(isamAppliance1, isamAppliance2):
    """
    Compare resources between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    return tools.json_compare(ret_obj1, ret_obj2)



import logging

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/rsyslog_forwarder/facility_names"
requires_modules = None
requires_version = "9.0.2.1"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the remote syslog forwarding facility names
    """
    return isamAppliance.invoke_get(
        "Retrieve the remote syslog forwarding facility names", uri, requires_modules=requires_modules,
        requires_version=requires_version)



import logging

logger = logging.getLogger(__name__)

uri = "/wga/common_configuration"
requires_modules = ["wga"]
requires_version = None


def get_all(isamAppliance, reverseproxy_id, check_mode=False, force=False):
    """
    Retrieving all common configuration for reverse proxy
    """
    return isamAppliance.invoke_get("Retrieving the all common configuration for reverse proxy",
                                    "{0}/{1}".format(uri, reverseproxy_id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, reverseproxy_id, configuration_id, check_mode=False, force=False):
    """
    Retrieving value of specified configuration id from reverse proxy common configuration
    """
    ret_object = get_all(isamAppliance, reverseproxy_id, check_mode, force)

    logger.debug("Looking for {0} value in reverse proxy {1}".format(configuration_id, reverseproxy_id))

    ret_obj = isamAppliance.create_return_object()

    try:
        ret_obj['data'] = ret_object['data'][configuration_id]
    except:
        logger.error("Invalid configuration_id: {0}".format(configuration_id))

    return ret_obj



import json
import logging
import time

logger = logging.getLogger(__name__)


def reboot(isdsAppliance, check_mode=False, force=False):
    """
    Reboot the appliance
    """
    if check_mode is True:
        return isdsAppliance.create_return_object(changed=True)
    else:
        return isdsAppliance.invoke_post("Rebooting appliance",
                                         "/diagnostics/restart_shutdown/reboot",
                                         {})


def shutdown(isdsAppliance, check_mode=False, force=False):
    """
    Shutdown the appliance
    """
    if check_mode is True:
        return isdsAppliance.create_return_object(changed=True)
    else:
        return isdsAppliance.invoke_post("Shutting down appliance",
                                         "/diagnostics/restart_shutdown/shutdown",
                                         {})


def _changes_available(isdsAppliance):
    """
    Check for pending changes on the appliance
    :param isdsAppliance:
    :return:
    """
    #    changes = isdsAppliance.invoke_get("Get pending changes",
    #                                       "/pending_changes")
    #    logger.debug("Pending changes on appliance are:")
    #    logger.debug(changes['data'])
    #
    #    change_count = isdsAppliance.invoke_get("Get pending changes count",
    #                                            "/pending_changes/count")
    #    logger.debug("Pending change count on appliance is:")
    #    logger.debug(change_count['data'])
    #
    #    if change_count['data']['count'] > 0 or len(changes['data']['changes']) > 0:
    #        logger.info("pending changes found")
    #        return True
    #    else:
    #        return False

    return True


def commit(isdsAppliance, check_mode=False, force=False):
    """
    Commit the current pending changes.
    """
    if force is True or _changes_available(isdsAppliance) is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_get("Committing the changes",
                                            "/pending_changes/deploy",
                                            {})

    return isdsAppliance.create_return_object()


def commit_and_restart(isdsAppliance, check_mode=False, force=False):
    """
    Commit and Restart LMI
    :param isdsAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _changes_available(isdsAppliance) is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_post("Commit and Restart",
                                             "/restarts/commit_and_restart",
                                             {})
    return isdsAppliance.create_return_object()


def rollback(isdsAppliance, check_mode=False, force=False):
    """
    Rollback the current pending changes.
    """
    if force is True or _changes_available(isdsAppliance) is True:
        if check_mode is True:
            return isdsAppliance.create_return_object(changed=True)
        else:
            return isdsAppliance.invoke_get("Rollback the changes",
                                            "/pending_changes/forget")

    return isdsAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools
from ibmsecurity.isam.aac import attributes as attrib
import json

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/risk/profiles/"
requires_modules = ["mga"]
requires_version = None


def get_all(isamAppliance, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of risk profiles
    """
    return isamAppliance.invoke_get("Retrieve a list of Risk Profiles",
                                    "{0}/{1}".format(uri, tools.create_query_string(filter=filter, sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific risk profile
    """
    warnings = []
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        warnings.append("Risk Profile {0} had no match, skipping retrieval.".format(name))
        return isamAppliance.create_return_object(changed=False, warnings=warnings)
    else:
        return _get(isamAppliance, id)


def _get(isamAppliance, id):
    return isamAppliance.invoke_get("Retrieve a specific Risk Profile",
                                    "{0}/{1}".format(uri, id))


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search risk profile id by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info("Found Risk Profile {0} id: {1}".format(name, obj['id']))
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, active, description=None, attributes=None, predefined=False, check_mode=False,
        force=False):
    """
    Creating or Modifying a Risk Profile
    """
    warnings = []
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know risk profile does not exist
        logger.info("Risk Profile {0} had no match, requesting to add new one.".format(name))
        return add(isamAppliance, name, active, description, attributes, predefined, check_mode, True)
    else:
        # Update request
        logger.info("Risk Profile {0} exists, requesting to update.".format(name))
        return update(isamAppliance, name, active, description, attributes, predefined, check_mode, force)


def add(isamAppliance, name, active, description=None, attributes=None, predefined=False, check_mode=False,
        force=False):
    """
    Create a new Risk Profile
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "name": name,
                "active": active,
                "predefined": predefined
            }
            if attributes is not None:
                logger.info("Check for empty attributes")
                # add loop through the attributes
                # find missing attributeID and go look them up
                for attv in attributes:
                    logger.debug("Checking {0}".format(attv['name']))
                    if 'attributeID' not in attv:
                        # lookup the attributeID from the name
                        logger.info(
                            "AttributeID for {0} was empty".format( attv['name']))
                        ret_obj = attrib.search(isamAppliance, attv['name'])
                        artifact_id = ret_obj['data']
                        if artifact_id == {}:
                            logger.debug("Attribute {0} had no match, skipping retrieval.".format( attv['name'] ))
                        else:
                            attv['attributeID'] = artifact_id
                json_data['attributes'] = attributes
            if description is not None:
                json_data['description'] = description
            return isamAppliance.invoke_post(
                "Create a new Risk Profile", uri, json_data)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, active, description=None, attributes=None, predefined=False, check_mode=False,
           force=False):
    """
    Update a specified Risk Profile
    """
    # add loop through the attributes
    # find missing attributeID and go look them up
    if attributes is not None:
       for attv in attributes:
           if 'attributeID' not in attv:
               # lookup the attributeID from the name
               logger.debug("AttributeID for {0} was empty".format(attv['name']))
               ret_obj = attrib.search(isamAppliance, attv['name'])
               artifact_id = ret_obj['data']
               if artifact_id == {}:
                   logger.debug("Attribute {0} had no match, skipping retrieval.".format(name))
               else:
                   attv['attributeID'] = artifact_id

    id, update_required, json_data = _check(isamAppliance, name, active, description, attributes, predefined)
    if id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", "Cannot update data for unknown Risk Profile: {0}".format(name))

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified Risk Profile",
                "{0}/{1}".format(uri, id), json_data)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, active, description, attributes, predefined):
    """
    Check and return True if update needed
    """
    update_required = False
    ret_obj = get(isamAppliance, name)
    json_data = {}
    if ret_obj['data'] == {}:
        logger.warning("Risk Profile not found, returning no update required.")
        return None, update_required, json_data
    else:
        if ret_obj['data']['predefined'] is True:
            logger.warning("Predefined Risk Profiles can NOT be updated, returning no update required.")
            return ret_obj['data']['id'], update_required, {}
        else:
            json_data = {
                "name": name,
                "active": active,
                "predefined": predefined
            }
            if attributes is not None:
                json_data['attributes'] = attributes
            else:
                del ret_obj['data']['attributes']
            if description is not None:
                json_data['description'] = description
            else:
                del ret_obj['data']['description']

            id = ret_obj['data']['id']
            del ret_obj['data']['id']

            sorted_json_data = json.dumps(json_data, skipkeys=True, sort_keys=True)
            logger.debug("Sorted input: {0}".format(sorted_json_data))
            sorted_ret_obj = json.dumps(ret_obj['data'], skipkeys=True, sort_keys=True)
            logger.debug("Sorted existing data: {0}".format(sorted_ret_obj))
            if sorted_ret_obj != sorted_json_data:
                logger.info("Changes detected, update needed.")
                update_required = True

    return id, update_required, json_data


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete a specified Risk Profile
    """
    ret_obj = get(isamAppliance, name, check_mode=check_mode, force=force)
    prof_id = ret_obj['data']['id']
    predef = ret_obj['data']['predefined']
    if prof_id == {}:
        logger.info("Risk Profile {0} not found, skipping delete.".format(name))
    else:
        if predef is True:
            logger.info("Risk Profile {0} is predefined, skipping delete.".format(name))
        else:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_delete(
                    "Delete a Risk Profile",
                    "{0}/{1}".format(uri, prof_id))

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Risk Profiles between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)
    for obj in ret_obj1['data']:
        del obj['id']
    for obj in ret_obj2['data']:
        del obj['id']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id'])



"""
Explorative implement for un-documented ISVG RESTAPI (not functional as of now)
Preserved here for the future in case if support (RFE)
"""
import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isvgAppliance):
    """
    Get Set up complete status
    """
    return isvgAppliance.invoke_get("Get Setup Complete Settings",
                                    "/setup_complete")


def set(isvgAppliance, check_mode=False, force=False):
    """
    Setup complete
    """
    if force is False:
        ret_obj = get(isvgAppliance)

    if force is True or ret_obj['data'].get('configured') is not True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_put(
                "Setup complete",
                "/setup_complete",
                {})

    return isvgAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools
import os.path
from ibmsecurity.utilities.tools import files_same, get_random_temp_dir
import shutil

logger = logging.getLogger(__name__)

requires_modules = ["mga", "federation"]
requires_version = None


def get(isamAppliance, path, check_mode=False, force=False):
    """
    Retrieving the current runtime template files directory contents
    """
    return isamAppliance.invoke_get("Retrieving the current runtime template files directory contents",
                                    "/mga/template_files/{0}".format(path), requires_modules=requires_modules,
                                    requires_version=requires_version)


def _check(isamAppliance, path, name):
    ret_obj = get(isamAppliance, path)

    for obj in ret_obj['data']['contents']:
        if obj['name'] == name and obj['type'] == 'File':
            logger.info("File .{0}".format(obj['name']))
            return True

    return None


def create(isamAppliance, path, name, contents=None, check_mode=False, force=False):
    """
    Creating a file in the runtime template files directory

    :param isamAppliance:
    :param path:
    :param name:
    :param contents:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file != None:
        warnings.append("File {0} exists in path {1}. Ignoring create.".format(name, path))

    if force is True or check_file == None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Creating a file in the runtime template files directory",
                "/mga/template_files/{0}".format(path),
                {
                    'file_name': name,
                    'type': 'file',
                    'contents': contents
                }, requires_modules=requires_modules,
                requires_version=requires_version)
    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, path, name, contents=None, check_mode=False, force=False):
    """
    Update a file in the runtime template files directory
    :param isamAppliance:
    :param path:
    :param name:
    :param contents:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file == None:
        warnings.append("File {0} does not exists in path {1}. Ignoring update.".format(name, path))
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or check_file is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            if contents is not None:
                return isamAppliance.invoke_put_files(
                    "Update a file in the runtime template files directory",
                    "/mga/template_files/{0}/{1}".format(path, name),
                    {
                        'contents': contents,
                        'type': 'file'
                    }, requires_modules=requires_modules,
                    requires_version=requires_version)
            else:
                warnings.append("Either contents or filename parameter need to be provided. Skipping update request.")
                return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, path, name, check_mode=False, force=False):
    """
    Deleting a file in the runtime template files directory

    :param isamAppliance:
    :param path:
    :param name:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file == None:
        warnings.append("File {0} does not exists in path {1}. Ignoring delete.".format(name, path))
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or check_file != None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a file in the runtime template files directory",
                "/mga/template_files/{0}/{1}".format(path, name), requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def rename(isamAppliance, path, name, new_name, check_mode=False, force=False):
    """
    Deleting a file in the runtime template files directory

    :param isamAppliance:
    :param path:
    :param name:
    :param new_name:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file == None:
        warnings.append("File {0} does not exists in path {1}. Ignoring rename.".format(name, path))
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or check_file != None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Renaming a file in the runtime template files directory",
                "/mga/template_files/{0}/{1}".format(path, name),
                {
                    'new_name': new_name,
                    'type': 'file'
                }, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def export_file(isamAppliance, path, name, filename, check_mode=False, force=False):
    """
    Exporting a file from the runtime template files directory

    :param isamAppliance:
    :param path:
    :param name:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file == None:
        warnings.append("File {0} does not exists in path {1}. Ignoring export.".format(name, path))
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or check_file != None:
        if check_mode is False:
            return isamAppliance.invoke_get_file(
                "Exporting a file from the runtime template files directory",
                "/mga/template_files/{0}/{1}?type=File&export=true".format(path, name), filename,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check_import(isamAppliance, path, name, filename, check_mode=False):
    """
    Checks if file on the Appliance (name) exists and if so, whether it is different from filename
    :param isamAppliance:
    :param path:
    :param name:
    :param filename:
    :return:
    """
    tmpdir = get_random_temp_dir()
    tmp_original_file = os.path.join(tmpdir, os.path.basename(name))
    if _check(isamAppliance, path, name):
        export_file(isamAppliance, path, name, tmp_original_file, check_mode=False, force=True)
        logger.debug("file already exists on appliance")
        if files_same(tmp_original_file, filename):
            logger.debug("files are the same, so we don't want to do anything")
            shutil.rmtree(tmpdir)
            return False
        else:
            logger.debug("files are different, so we delete existing file in preparation for import")
            delete(isamAppliance, path, name, check_mode=check_mode, force=True)
            shutil.rmtree(tmpdir)
            return True
    else:
        logger.debug("file does not exist on appliance, so we'll want to import")
        shutil.rmtree(tmpdir)
        return True


def import_file(isamAppliance, path, name, filename, check_mode=False, force=False):
    """
    Importing a file in the runtime template files directory.
    """
    warnings = []
    check_file = _check(isamAppliance, path, name)
    if check_file != None and force == False:
        warnings.append("File {0} exists in path {1}.".format(name, path))

    if force is True or _check_import(isamAppliance, path, name, filename, check_mode=check_mode):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a file in the runtime template files directory",
                "/mga/template_files/{0}/{1}".format(path, name),
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    'type': 'file',
                    'force': force
                }, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)



import logging
from .user import User


class ISDSApplianceUser(User):
    super_user = "admin"  # Use name that will work regardless of Management Authentication setting

    def __init__(self, password, username=None):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating a user')

        if username is None:
            User.__init__(self, password, self.super_user)
            self.logger.debug('Creating a super-user')
        else:
            User.__init__(self, password, username)



import json
import logging
import time
import ibmsecurity.isam.base.lmi

logger = logging.getLogger(__name__)
requires_model="Appliance"


def reboot(isamAppliance, check_mode=False, force=False):
    """
    Restart the appliance
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post("Restart the appliance",
                                         "/diagnostics/restart_shutdown/reboot",
                                         {}, requires_model=requires_model)


def shutdown(isamAppliance, check_mode=False, force=False):
    """
    Shutdown the appliance
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_post("Shutting down appliance",
                                         "/diagnostics/restart_shutdown/shutdown",
                                         {}, requires_model=requires_model)


def _changes_available(isamAppliance):
    """
    Check for pending changes on the appliance
    :param isamAppliance:
    :return:
    """
    changes = isamAppliance.invoke_get("Get pending changes",
                                       "/isam/pending_changes")
    logger.debug("Pending changes on appliance are:")
    logger.debug(changes['data'])
    logger.debug(changes)

    change_count = isamAppliance.invoke_get("Get pending changes count",
                                            "/isam/pending_changes/count")
    logger.debug("Pending change count on appliance is:")
    logger.debug(change_count['data'])
    logger.debug(change_count)

    if change_count['data']['count'] > 0 or len(changes['data']['changes']) > 0:
        logger.info("pending changes found")
        return True
    else:
        return False


def commit(isamAppliance, check_mode=False, force=False):
    """
    Commit the current pending changes.
    """
    if force is True or _changes_available(isamAppliance) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Committing the changes",
                                            "/isam/pending_changes",
                                            {})

    return isamAppliance.create_return_object()


def commit_and_restart(isamAppliance, check_mode=False, force=False):
    """
    Commit and Restart
    :param isamAppliance:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _changes_available(isamAppliance) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Commit and Restart",
                                             "/restarts/commit_and_restart",
                                             {})
    return isamAppliance.create_return_object()


def reboot_and_wait(isamAppliance, wait_time=300, check_freq=5, check_mode=False, force=False):
    """
    Reboot and wait
    :param isamAppliance:
    :param wait_time:
    :param check_freq:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        firmware = ibmsecurity.isam.base.firmware.get(isamAppliance, check_mode=check_mode, force=force)

        ret_obj = reboot(isamAppliance)

        if ret_obj['rc'] == 0:
            sec = 0

            # Now check if it is up and running
            while 1:
                ret_obj = ibmsecurity.isam.base.firmware.get(isamAppliance, check_mode=check_mode, force=force,
                                                             ignore_error=True)

                # check partition last_boot time
                if ret_obj['rc'] == 0 and isinstance(ret_obj['data'], list) and len(ret_obj['data']) > 0 and \
                        (('last_boot' in ret_obj['data'][0] and ret_obj['data'][0]['last_boot'] != firmware['data'][0][
                            'last_boot'] and ret_obj['data'][0]['active'] == True) or (
                                 'last_boot' in ret_obj['data'][1] and ret_obj['data'][1]['last_boot'] !=
                                 firmware['data'][1]['last_boot'] and ret_obj['data'][1]['active'] == True)):
                    logger.info("Server is responding and has a different boot time!")
                    return isamAppliance.create_return_object(warnings=warnings)
                else:
                    time.sleep(check_freq)
                    sec += check_freq
                    logger.debug(
                        "Server is not responding yet. Waited for {0} secs, next check in {1} secs.".format(sec,
                                                                                                            check_freq))

                if sec >= wait_time:
                    warnings.append("Server reboot not detected or completed, exiting... after {0} seconds".format(sec))
                    break

    return isamAppliance.create_return_object(warnings=warnings)


def commit_and_restart_and_wait(isamAppliance, wait_time=300, check_freq=5, check_mode=False, force=False):
    """
    Restart LMI after commit
    :param isamAppliance:
    :param wait_time:
    :param check_freq:
    :param check_mode:
    :param force:
    :return:
    """
    warnings = []
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        lmi = ibmsecurity.isam.base.lmi.get(isamAppliance, check_mode=check_mode, force=force)

        ret_obj = commit_and_restart(isamAppliance)

        if ret_obj['rc'] == 0:
            sec = 0

            # Now check if it is up and running
            while 1:
                ret_obj = ibmsecurity.isam.base.lmi.get(isamAppliance, check_mode=check_mode, force=force)

                # check partition last_boot time
                if ret_obj['rc'] == 0 and isinstance(ret_obj['data'], list) and len(ret_obj['data']) > 0 and \
                        ('start_time' in ret_obj['data'][0] and ret_obj['data'][0]['start_time'] != lmi['data'][0][
                            'start_time']):
                    logger.info("LMI is responding and has a different start time!")
                    return isamAppliance.create_return_object(warnings=warnings)
                else:
                    time.sleep(check_freq)
                    sec += check_freq
                    logger.debug("LMI is not responding yet. Waited for {0} secs, next check in {1} secs.".format(sec,
                                                                                                                  check_freq))

                if sec >= wait_time:
                    warnings.append(
                        "The LMI restart not detected or completed, exiting... after {0} seconds".format(sec))
                    break

    return isamAppliance.create_return_object(warnings=warnings)


def rollback(isamAppliance, check_mode=False, force=False):
    """
    Rollback the current pending changes.
    """
    if force is True or _changes_available(isamAppliance) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete("Rollback the changes",
                                               "/isam/pending_changes")

    return isamAppliance.create_return_object()
